"""Calculation of flip ratio."""
import numpy

from .function_1_algebra import calc_modulus_sq_by_complex_vector,\
    calc_scalar_product_by_complex_vectors

from .function_3_extinction import calc_extinction_2


def calc_f_plus_sq(
        f_nucl: complex, f_mag_perp_vert: complex, flag_f_nucl: bool = False,
        flag_f_mag_perp_vert: bool = False):
    r"""F_{+}^{2} = (F_N + F_{M, \perp, z})^{2}."""
    flag_derivatives = any([flag_f_nucl, flag_f_mag_perp_vert])
    f_plus = f_nucl + f_mag_perp_vert
    f_plus_sq = (f_plus*f_plus.conjugate()).real
    dder = {}
    if flag_derivatives:
        if flag_f_nucl:
            der_f_plus_sq__f_nucl = 2.*f_plus*(1.+0.*f_nucl)
            dder["der__f_nucl_re"] = der_f_plus_sq__f_nucl.real
            dder["der__f_nucl_im"] = der_f_plus_sq__f_nucl.imag
        if flag_f_mag_perp_vert:
            der_f_plus_sq__f_mag_perp_vert = 2.*f_plus*(1.+0.*f_mag_perp_vert)
            dder["der__f_mag_perp_vert_re"] = \
                der_f_plus_sq__f_mag_perp_vert.real
            dder["der__f_mag_perp_vert_im"] = \
                der_f_plus_sq__f_mag_perp_vert.imag
    return f_plus_sq, dder


def calc_f_minus_sq(
        f_nucl: complex, f_mag_perp_vert: complex, flag_f_nucl: bool = False,
        flag_f_mag_perp_vert: bool = False):
    r"""F_{-}^{2} = (F_N - F_{M, \perp, z})^{2}."""
    flag_derivatives = any([flag_f_nucl, flag_f_mag_perp_vert])
    f_minus = f_nucl - f_mag_perp_vert
    f_minus_sq = (f_minus*f_minus.conjugate()).real
    dder = {}
    if flag_derivatives:
        if flag_f_nucl:
            der_f_minus_sq__f_nucl = 2.*f_minus*(1.+0.*f_nucl)
            dder["der__f_nucl_re"] = der_f_minus_sq__f_nucl.real
            dder["der__f_nucl_im"] = der_f_minus_sq__f_nucl.imag
        if flag_f_mag_perp_vert:
            der_f_minus_sq__f_mag_perp_vert = 2.*f_minus*(
                1.+0.*f_mag_perp_vert)
            dder["der__f_mag_perp_vert_re"] = \
                der_f_minus_sq__f_mag_perp_vert.real
            dder["der__f_mag_perp_vert_im"] = \
                der_f_minus_sq__f_mag_perp_vert.imag
    return f_minus_sq, dder


def calc_flip_ratio(
        f_nucl: float, f_mag_perp: tuple, e_up: tuple,
        beam_polarization: float, flipper_efficiency: float, radius: float,
        mosaicity: float, model_extinction: str, volume_unit_cell: float,
        sthovl: float, wavelength: float,
        flag_f_nucl: bool = False,
        flag_f_mag_perp: bool = False, flag_e_up: bool = False,
        flag_beam_polarization: bool = False,
        flag_flipper_efficiency: bool = False, flag_radius: bool = False,
        flag_mosaicity: bool = False, flag_volume_unit_cell: bool = False,
        flag_sthovl: bool = False, flag_volume_wavelength: bool = False):
    """Calculate flip ratio.

    Calculate flip ratio by given nuclear structure factor and
    perpendicular component of magnetic structure factor
    (in 10**-12 cm)

    cell (only for extinction correction)
    f_nucl [hkl]
    f_m_perp [hkl]
    delta_f_nucl [hkl, parameters]
    delta_f_m_perp [hkl, parameters]
    """
    flag_derivatives = any([
        flag_f_nucl, flag_f_mag_perp, flag_e_up, flag_beam_polarization,
        flag_flipper_efficiency, flag_radius, flag_mosaicity,
        flag_volume_unit_cell, flag_sthovl, flag_volume_wavelength])

    np_b_p = numpy.array(beam_polarization, dtype=float)
    np_f_e = numpy.array(flipper_efficiency, dtype=float)
    p_u = np_b_p
    p_d = (2. * np_f_e - 1.) * p_u

    part_u_u, part_u_d = 0.5*(1. + p_u), 0.5*(1. - p_u)
    part_d_u, part_d_d = 0.5*(1. - p_d), 0.5*(1. + p_d)

    f_mag_perp_vert, dder_f_mag_perp_vert = \
        calc_scalar_product_by_complex_vectors(
            f_mag_perp, e_up, flag_vector_1=flag_f_mag_perp,
            flag_vector_2=flag_e_up)

    if flag_f_mag_perp:
        der_f_mag_perp_vert__f_mag_perp_re = \
            dder_f_mag_perp_vert["der__vector_1_re"]
        der_f_mag_perp_vert__f_mag_perp_im = \
            dder_f_mag_perp_vert["der__vector_1_im"]

    if flag_e_up:
        der_f_mag_perp_vert__e_up = dder_f_mag_perp_vert["der__vector_2_re"]

    flag_f_mag_perp_vert = any([flag_f_mag_perp, flag_e_up])

    f_plus_sq, dder_f_plus_sq = calc_f_plus_sq(
        f_nucl, f_mag_perp_vert, flag_f_nucl, flag_f_mag_perp_vert)
    f_minus_sq, dder_f_minus_sq = calc_f_minus_sq(
        f_nucl, f_mag_perp_vert, flag_f_nucl, flag_f_mag_perp_vert)

    if flag_f_nucl:
        der_f_plus_sq__f_nucl_re = \
            dder_f_plus_sq["der__f_nucl_re"]
        der_f_plus_sq__f_nucl_im = \
            dder_f_plus_sq["der__f_nucl_im"]
        der_f_minus_sq__f_nucl_re = \
            dder_f_minus_sq["der__f_nucl_re"]
        der_f_minus_sq__f_nucl_im = \
            dder_f_minus_sq["der__f_nucl_im"]
    if flag_f_mag_perp_vert:
        der_f_plus_sq__f_mag_perp_vert_re = \
            dder_f_plus_sq["der__f_mag_perp_vert_re"]
        der_f_plus_sq__f_mag_perp_vert_im = \
            dder_f_plus_sq["der__f_mag_perp_vert_im"]
        der_f_minus_sq__f_mag_perp_vert_re = \
            dder_f_minus_sq["der__f_mag_perp_vert_re"]
        der_f_minus_sq__f_mag_perp_vert_im = \
            dder_f_minus_sq["der__f_mag_perp_vert_im"]

    f_mag_perp_sq, dder_f_mag_perp_sq = calc_modulus_sq_by_complex_vector(
        f_mag_perp, flag_f_mag_perp)

    if flag_f_mag_perp:
        der_f_mag_perp_sq__f_mag_perp_re = \
            dder_f_mag_perp_sq["der__vector_1_re"]
        der_f_mag_perp_sq__f_mag_perp_im = \
            dder_f_mag_perp_sq["der__vector_1_im"]

    f_mag_perp_vert_sq = (f_mag_perp_vert*f_mag_perp_vert.conjugate()).real
    if flag_f_mag_perp_vert:
        der_f_mag_perp_vert_sq__f_mag_perp_vert_re = 2.*f_mag_perp_vert.real
        der_f_mag_perp_vert_sq__f_mag_perp_vert_im = 2.*f_mag_perp_vert.imag

    f_mag_sq = f_mag_perp_sq - f_mag_perp_vert_sq

    if any([flag_f_mag_perp, flag_f_mag_perp_vert]):
        der_f_mag_sq__f_mag_perp_re = der_f_mag_perp_sq__f_mag_perp_re - \
            der_f_mag_perp_vert_sq__f_mag_perp_vert_re * \
            der_f_mag_perp_vert__f_mag_perp_re.real + \
            der_f_mag_perp_vert_sq__f_mag_perp_vert_im * \
            der_f_mag_perp_vert__f_mag_perp_re.imag

        der_f_mag_sq__f_mag_perp_im = der_f_mag_perp_sq__f_mag_perp_im - \
            der_f_mag_perp_vert_sq__f_mag_perp_vert_re * \
            der_f_mag_perp_vert__f_mag_perp_im.real + \
            der_f_mag_perp_vert_sq__f_mag_perp_vert_im * \
            der_f_mag_perp_vert__f_mag_perp_im.imag

    # extinction correction
    y_plus, dder_y_plus = calc_extinction_2(
        radius, mosaicity, model_extinction, f_plus_sq, volume_unit_cell,
        sthovl, wavelength)
    y_minus, dder_y_minus = calc_extinction_2(
        radius, mosaicity, model_extinction, f_minus_sq, volume_unit_cell,
        sthovl, wavelength)
    y_mag, dder_y_mag = calc_extinction_2(
        radius, mosaicity, model_extinction, f_mag_sq, volume_unit_cell,
        sthovl, wavelength)
    if flag_derivatives:
        der_y_plus__f_plus_sq = dder_y_plus["der_yext__f_sq"]
        der_y_minus__f_minus_sq = dder_y_minus["der_yext__f_sq"]
        der_y_mag__f_mag_sq = dder_y_mag["der_yext__f_sq"]

    int_up = part_u_u*y_plus*f_plus_sq + part_u_d*y_minus*f_minus_sq + \
        y_mag*f_mag_sq
    int_down = part_d_u*y_plus*f_plus_sq + part_d_d*y_minus*f_minus_sq + \
        y_mag*f_mag_sq

    if flag_derivatives:
        der_int_up__f_plus_sq = part_u_u*(1.+0.*f_plus_sq)*y_plus
        der_int_up__f_minus_sq = part_u_d*(1.+0.*f_minus_sq)*y_minus
        der_int_up__f_mag_sq = (1.+0.*f_mag_sq)*y_mag
        der_int_up__y_plus = part_u_u*f_plus_sq*(1.+0.*y_plus)
        der_int_up__y_minus = part_u_d*f_minus_sq*(1.+0.*y_minus)
        der_int_up__y_mag = f_mag_sq*(1.+0.*y_mag)

        der_int_down__f_plus_sq = part_d_u*(1.+0.*f_plus_sq)*y_plus
        der_int_down__f_minus_sq = part_d_d*(1.+0.*f_minus_sq)*y_minus
        der_int_down__f_mag_sq = der_int_up__f_mag_sq
        der_int_down__y_plus = part_d_u*f_plus_sq*(1.+0.*y_plus)
        der_int_down__y_minus = part_d_d*f_minus_sq*(1.+0.*y_minus)
        der_int_down__y_mag = der_int_up__y_mag

    flip_ratio = int_up / int_down

    dder = {}
    # derivatives
    if flag_derivatives:
        der_flip_ratio__int_up = (1.+0* int_up) / int_down
        der_flip_ratio__int_down = - flip_ratio / int_down

    if flag_beam_polarization:
        tt = y_plus*f_plus_sq - y_minus*f_minus_sq
        der_int_up__beam_polarization = 0.5*tt*(1.+0.*np_b_p)
        der_int_down__beam_polarization = -0.5*(2.* np_f_e - 1.)*tt*(
            1.+0.*np_b_p)
        der_flip_ratio__beam_polarization = der_flip_ratio__int_up * \
            der_int_up__beam_polarization + der_flip_ratio__int_down * \
            der_int_down__beam_polarization
        dder["der_flip_ratio__beam_polarization"] = \
            der_flip_ratio__beam_polarization

    if flag_flipper_efficiency:
        tt = y_plus*f_plus_sq - y_minus*f_minus_sq
        der_int_down__flipper_efficiency = -1.0*(0.* np_f_e + 1.)*tt*np_b_p
        der_flip_ratio__flipper_efficiency = der_flip_ratio__int_down * \
            der_int_down__flipper_efficiency
        dder["der_flip_ratio__flipper_efficiency"] = \
            der_flip_ratio__flipper_efficiency

    if flag_f_mag_perp:
        der_f_plus_sq__f_mag_perp_re =  der_f_plus_sq__f_mag_perp_vert_re * \
            der_f_mag_perp_vert__f_mag_perp_re.real + \
            der_f_plus_sq__f_mag_perp_vert_im * \
            der_f_mag_perp_vert__f_mag_perp_re.imag

        der_f_minus_sq__f_mag_perp_re =  der_f_minus_sq__f_mag_perp_vert_re * \
            der_f_mag_perp_vert__f_mag_perp_re.real + \
            der_f_minus_sq__f_mag_perp_vert_im * \
            der_f_mag_perp_vert__f_mag_perp_re.imag

        der_int_up__f_mag_perp_re = der_int_up__f_plus_sq * \
            der_f_plus_sq__f_mag_perp_re + der_int_up__f_minus_sq * \
            der_f_minus_sq__f_mag_perp_re + der_int_up__y_plus * \
            der_y_plus__f_plus_sq * der_f_plus_sq__f_mag_perp_re + \
            der_int_up__y_minus * der_y_minus__f_minus_sq * \
            der_f_minus_sq__f_mag_perp_re + der_int_up__y_mag * \
            der_y_mag__f_mag_sq * der_f_mag_sq__f_mag_perp_re

        der_f_plus_sq__f_mag_perp_im =  der_f_plus_sq__f_mag_perp_vert_re * \
            der_f_mag_perp_vert__f_mag_perp_im.real + \
            der_f_plus_sq__f_mag_perp_vert_im * \
            der_f_mag_perp_vert__f_mag_perp_im.imag

        der_f_minus_sq__f_mag_perp_im =  der_f_minus_sq__f_mag_perp_vert_re * \
            der_f_mag_perp_vert__f_mag_perp_im.real + \
            der_f_minus_sq__f_mag_perp_vert_im * \
            der_f_mag_perp_vert__f_mag_perp_im.imag

        der_int_up__f_mag_perp_im = der_int_up__f_plus_sq * \
            der_f_plus_sq__f_mag_perp_im + der_int_up__f_minus_sq * \
            der_f_minus_sq__f_mag_perp_im + der_int_up__y_plus * \
            der_y_plus__f_plus_sq * der_f_plus_sq__f_mag_perp_im + \
            der_int_up__y_minus * der_y_minus__f_minus_sq * \
            der_f_minus_sq__f_mag_perp_im + der_int_up__y_mag * \
            der_y_mag__f_mag_sq * der_f_mag_sq__f_mag_perp_im

        der_int_down__f_mag_perp_re = der_int_down__f_plus_sq * \
            der_f_plus_sq__f_mag_perp_re + der_int_down__f_minus_sq * \
            der_f_minus_sq__f_mag_perp_re + der_int_down__y_plus * \
            der_y_plus__f_plus_sq * der_f_plus_sq__f_mag_perp_re + \
            der_int_down__y_minus * der_y_minus__f_minus_sq * \
            der_f_minus_sq__f_mag_perp_re + der_int_down__y_mag * \
            der_y_mag__f_mag_sq * der_f_mag_sq__f_mag_perp_re

        der_int_down__f_mag_perp_im = der_int_down__f_plus_sq * \
            der_f_plus_sq__f_mag_perp_im + der_int_down__f_minus_sq * \
            der_f_minus_sq__f_mag_perp_im + der_int_down__y_plus * \
            der_y_plus__f_plus_sq * der_f_plus_sq__f_mag_perp_im + \
            der_int_down__y_minus * der_y_minus__f_minus_sq * \
            der_f_minus_sq__f_mag_perp_im + der_int_down__y_mag * \
            der_y_mag__f_mag_sq * der_f_mag_sq__f_mag_perp_im

        der_flip_ratio__f_mag_perp_re = der_flip_ratio__int_up * \
            der_int_up__f_mag_perp_re + der_flip_ratio__int_down * \
            der_int_down__f_mag_perp_re

        der_flip_ratio__f_mag_perp_im = der_flip_ratio__int_up * \
            der_int_up__f_mag_perp_im + der_flip_ratio__int_down * \
            der_int_down__f_mag_perp_im

        dder["der_flip_ratio__f_mag_perp_re"] = der_flip_ratio__f_mag_perp_re
        dder["der_flip_ratio__f_mag_perp_im"] = der_flip_ratio__f_mag_perp_im

    return flip_ratio, dder




"""
Base crystallographic functions.
"""
import numpy
from cryspy.A_functions_base.function_1_matrices import calc_mRmCmRT, \
    calc_product_matrix_vector, scalar_product

from .unit_cell import calc_sthovl_by_unit_cell_parameters


def calc_cos_ang(cell, h_1, k_1, l_1, h_2, k_2, l_2):
    """Calculate directed cosines."""
    q_1_x, q_1_y, q_1_z = cell.calc_k_loc(h_1, k_1, l_1)
    q_2_x, q_2_y, q_2_z = cell.calc_k_loc(h_2, k_2, l_2)
    q_1_sq = q_1_x*q_1_x + q_1_y*q_1_y + q_1_z*q_1_z
    q_2_sq = q_2_x*q_2_x + q_2_y*q_2_y + q_2_z*q_2_z
    q_12 = q_1_x*q_2_x + q_1_y*q_2_y + q_1_z*q_2_z
    res = q_12/(q_1_sq*q_2_sq)**0.5
    res[res > 1.] = 1.
    return res


def calc_volume_uc_by_abc_cosines(a:float, b:float, c:float, 
        cos_alpha:float, cos_beta:float, cos_gamma:float):
    """
    Calculate the volume of unit cell defined as 
    a, b, c, cos(alpha), cos(beta), cos(gamma).
    """
    c_a_sq, c_b_sq = numpy.square(cos_alpha), numpy.square(cos_beta)
    c_g_sq = numpy.square(cos_gamma)
    v_sqrt = numpy.sqrt(1.-c_a_sq-c_b_sq-c_g_sq+2.*cos_alpha*cos_beta*cos_gamma)
    volume_uc = a*b*c*v_sqrt
    return volume_uc


def calc_volume_uc_by_abc_angles(a:float, b:float, c:float, 
                          alpha:float, beta:float, gamma:float):
    """
    Calculate the volume of unit cell defined as 
    a, b, c, alpha, beta, gamma.
    Angles should be given in radians.
    """
    c_a, c_b, c_g = numpy.cos(alpha), numpy.cos(beta), numpy.cos(gamma)
    volume_uc = calc_volume_uc_by_abc_cosines(a, b, c, c_a, c_b, c_g)
    return volume_uc


def calc_inverse_d_by_hkl_abc_cosines(h:float, k:float, l:float, 
        a:float, b:float, c:float, 
        cos_alpha:float, cos_beta:float, cos_gamma:float):
    """
    Calculate 1/d for given reflections h, k, l
    and unit cell parameters defined as a, b, c, cos(alpha), cos(beta), cos(gamma).
    """
    c_a_sq, c_b_sq = numpy.square(cos_alpha), numpy.square(cos_beta)
    c_g_sq = numpy.square(cos_gamma)
    s_a_sq, s_b_sq, s_g_sq = (1.-c_a_sq), (1.-c_b_sq), (1.-c_g_sq)

    A = (1.-c_a_sq-c_b_sq-c_g_sq+2.*cos_alpha*cos_beta*cos_gamma)

    B1 = s_a_sq*numpy.square(h*1./a)+\
         s_b_sq*numpy.square(k*1./b)+\
         s_g_sq*numpy.square(l*1./c)
    B2 = 2.*(k*l*(cos_alpha-cos_beta*cos_gamma))/(b*c)+\
         2.*(h*l*(cos_beta-cos_alpha*cos_gamma))/(a*c)+\
         2.*(h*k*(cos_gamma-cos_alpha*cos_beta))/(a*b)

    # B2 = 2.*(k*l*cos_alpha)/(b*c)+\
    #      2.*(h*l*cos_beta)/(a*c)+\
    #      2.*(h*k*cos_gamma)/(a*b)
    #it should be checked, I am not sure
    B = B1-B2
    inverse_d = numpy.sqrt(B*1./A)
    return inverse_d


def calc_inverse_d_by_hkl_abc_angles(
        h:float, k:float, l:float, a:float, b:float, c:float, alpha:float,
        beta:float, gamma:float):
    """
    Calculate 1/d for given reflections h, k, l
    and unit cell parameters defined as a, b, c, alpha, beta, gamma.
    Angles should be given in radians.
    """
    c_a, c_b, c_g = numpy.cos(alpha), numpy.cos(beta), numpy.cos(gamma)
    inverse_d = calc_inverse_d_by_hkl_abc_cosines(h, k, l, a, b, c, c_a, c_b, c_g)
    return inverse_d


def calc_sthovl_by_hkl_abc_cosines(
        h: float, k: float, l: float, a: float, b: float, c: float,
        cos_alpha: float, cos_beta: float, cos_gamma: float):
    """
    Calculate sin(theta)/lambda for given reflections h, k, l
    and unit cell parameters defined as a, b, c, cos(alpha), cos(beta), cos(gamma).
    """
    inv_d = calc_inverse_d_by_hkl_abc_cosines(h, k, l, a, b, c, 
                                              cos_alpha, cos_beta, cos_gamma)
    sthovl = 0.5*inv_d
    return sthovl


def calc_sthovl_by_hkl_abc_angles(
        h: float, k: float, l: float, a: float, b: float, c: float,
        alpha: float, beta: float, gamma: float):
    """
    Calculate sin(theta)/lambda for given reflections h, k, l
    and unit cell parameters defined as a, b, c, alpha, beta, gamma.
    Angles should be given in radians.
    """
    inv_d = calc_inverse_d_by_hkl_abc_angles(h, k, l, a, b, c, alpha, beta, gamma)
    sthovl = 0.5*inv_d
    return sthovl


def calc_phase_3d(hkl, r_ij, b_i, fract_xyz):
    """
Calculation of phases over 3 dimensions (hkl, points, symmetry):

fract_xyz.shape = (pointxs, 3)

r_ij, b_i are elements of symmetry

phase = exp(2\pi i (h*Rs*x + h*bs))

phase_3d: [hkl, points, symmetry]
    """
    h, k, l = hkl[0], hkl[1], hkl[2]
    x, y, z = fract_xyz[:, 0], fract_xyz[:, 1], fract_xyz[:, 2]
    r_11, r_12, r_13, r_21, r_22, r_23, r_31, r_32, r_33 = r_ij
    b_1, b_2, b_3 = b_i

    # [ind, symm]
    mr_2d_1, mr_2d_2, mr_2d_3 = calc_product_matrix_vector(
        (r_11[numpy.newaxis, :], r_12[numpy.newaxis, :], r_13[numpy.newaxis, :],
         r_21[numpy.newaxis, :], r_22[numpy.newaxis, :], r_23[numpy.newaxis, :],
         r_31[numpy.newaxis, :], r_32[numpy.newaxis, :], r_33[numpy.newaxis, :]), 
        (x[:, numpy.newaxis], y[:, numpy.newaxis], z[:, numpy.newaxis]))

    # [hkl, ind, symm]
    hrr_3d = scalar_product((h[:, numpy.newaxis, numpy.newaxis], 
                                     k[:, numpy.newaxis, numpy.newaxis], 
                                     l[:, numpy.newaxis, numpy.newaxis]),
        (mr_2d_1[numpy.newaxis, :, :], mr_2d_2[numpy.newaxis, :, :], mr_2d_3[numpy.newaxis, :, :]))

    # [hkl, symm]
    hb_2d = scalar_product((h[:, numpy.newaxis], k[:, numpy.newaxis], l[:, numpy.newaxis]), 
                           (b_1[numpy.newaxis, :], b_2[numpy.newaxis, :], b_3[numpy.newaxis, :]))
    phase_3d = numpy.exp(2.*numpy.pi*1j*(hrr_3d+hb_2d[:, numpy.newaxis,:]))
    return phase_3d


def calc_phase_by_hkl_xyz_rb(index_hkl, x, y, z, r_11, r_12, r_13, r_21, r_22,
                             r_23, r_31, r_32, r_33, b_1, b_2, b_3):
    h, k, l = index_hkl[0], index_hkl[1], index_hkl[2]
    np_h, np_x, np_r_11 = numpy.meshgrid(h, x, r_11, indexing="ij")
    np_k, np_y, np_r_22 = numpy.meshgrid(k, y, r_22, indexing="ij")
    np_l, np_z, np_r_33 = numpy.meshgrid(l, z, r_33, indexing="ij")

    np_r_12 = numpy.meshgrid(h, x, r_12, indexing="ij")[2]
    np_r_13 = numpy.meshgrid(k, y, r_13, indexing="ij")[2]
    np_r_23 = numpy.meshgrid(l, z, r_23, indexing="ij")[2]
    np_r_21 = numpy.meshgrid(h, x, r_21, indexing="ij")[2]
    np_r_31 = numpy.meshgrid(k, y, r_31, indexing="ij")[2]
    np_r_32 = numpy.meshgrid(l, z, r_32, indexing="ij")[2]
    np_b_1 = numpy.meshgrid(l, z, b_1, indexing="ij")[2]
    np_b_2 = numpy.meshgrid(l, z, b_2, indexing="ij")[2]
    np_b_3 = numpy.meshgrid(l, z, b_3, indexing="ij")[2]

    np_x_s = np_x*np_r_11 + np_y*np_r_12 + np_z*np_r_13 + np_b_1
    np_y_s = np_x*np_r_21 + np_y*np_r_22 + np_z*np_r_23 + np_b_2
    np_z_s = np_x*np_r_31 + np_y*np_r_32 + np_z*np_r_33 + np_b_3
    hh = (2*numpy.pi*1j*(np_h*np_x_s + np_k*np_y_s+ np_l*np_z_s)).astype(
        complex)
    phase_3d = numpy.exp(hh)
    return phase_3d


def calc_power_dwf_iso(b_iso, sthovl):
    """
isotropic harmonic Debye-Waller factor
    """
    sthovl_sq = sthovl**2
    b_iso_2d, sthovl_sq_2d = numpy.meshgrid(sthovl_sq, b_iso, indexing="ij")
    power_dwf_iso_2d = b_iso_2d*sthovl_sq_2d
    return power_dwf_iso_2d


def calc_power_dwf_aniso(index_hkl, beta, 
    r_11, r_12, r_13, r_21, r_22, r_23, r_31, r_32, r_33):
    """
anisotropic harmonic Debye-Waller factor

h,k,l is 1D (temporary solution)
    """
    h, k, l = index_hkl[0], index_hkl[1], index_hkl[2]
    b_11, b_22, b_33 = beta[:, 0], beta[:, 1], beta[:, 2]
    b_12, b_13, b_23 = beta[:, 3], beta[:, 4], beta[:, 5]

    np_h, np_b_11, np_r_11 = numpy.meshgrid(h, b_11, r_11, indexing="ij")
    np_k, np_b_22, np_r_22 = numpy.meshgrid(k, b_22, r_22, indexing="ij")
    np_l, np_b_33, np_r_33 = numpy.meshgrid(l, b_33, r_33, indexing="ij")
    np_h, np_b_12, np_r_12 = numpy.meshgrid(h, b_12, r_12, indexing="ij")
    np_h, np_b_13, np_r_13 = numpy.meshgrid(h, b_13, r_13, indexing="ij")
    np_h, np_b_23, np_r_23 = numpy.meshgrid(h, b_23, r_23, indexing="ij")
    np_r_21 = numpy.meshgrid(h, b_23, r_21, indexing="ij")[2]
    np_r_31 = numpy.meshgrid(h, b_23, r_31, indexing="ij")[2]
    np_r_32 = numpy.meshgrid(h, b_23, r_32, indexing="ij")[2]

    np_h_s = np_h*np_r_11 + np_k*np_r_21 + np_l*np_r_31
    np_k_s = np_h*np_r_12 + np_k*np_r_22 + np_l*np_r_32
    np_l_s = np_h*np_r_13 + np_k*np_r_23 + np_l*np_r_33

    power_dwf_aniso = (np_b_11*np_h_s**2 + np_b_22*np_k_s**2 +
                       np_b_33*np_l_s**2 + 2.*np_b_12*np_h_s*np_k_s +
                       2.*np_b_13*np_h_s*np_l_s + 2.*np_b_23*np_k_s*np_l_s)
    return power_dwf_aniso


def calc_dwf(cell, index_hkl, b_iso, beta, r_11, r_12, r_13, r_21, r_22, r_23,
             r_31, r_32, r_33):
    """Calculate Debye-Waller factor."""
    unit_cell_parameters = cell.get_unit_cell_parameters()
    sthovl = calc_sthovl_by_unit_cell_parameters(index_hkl, unit_cell_parameters, flag_unit_cell_parameters=False)[0]
    # dimensions (hkl, atoms in assymmetric unit cell)
    power_iso_2d = calc_power_dwf_iso(b_iso, sthovl)

    # dimensions (hkl, atoms in assymmetric unit cell, el.symmetry)
    power_aniso_3d = calc_power_dwf_aniso(
        index_hkl, beta, r_11, r_12, r_13, r_21, r_22, r_23, r_31, r_32, r_33)
    power_3d = power_iso_2d[:, :, numpy.newaxis] + power_aniso_3d
    dwf_3d = numpy.exp(-power_3d)
    return dwf_3d


def calc_form_factor_tensor_susceptibility(
        chi_11, chi_22, chi_33, chi_12, chi_13, chi_23, space_group_symop,
        form_factor, cell, h, k, l):
    """Give components of form factor tensor for susceptibility.

    fft_11, fft_12, fft_13
    fft_21, fft_22, fft_23
    fft_31, fft_32, fft_33

    in 3 dimension (hkl, atoms, symmetry elements).
    """
    ff = numpy.array(form_factor, dtype=float)
    sthovl = cell.calc_sthovl(h, k, l)
    # dimension (hkl, atoms)
    
    r_11 = numpy.array(space_group_symop.r_11, dtype=float)
    r_12 = numpy.array(space_group_symop.r_12, dtype=float)
    r_13 = numpy.array(space_group_symop.r_13, dtype=float)
    r_21 = numpy.array(space_group_symop.r_21, dtype=float)
    r_22 = numpy.array(space_group_symop.r_22, dtype=float)
    r_23 = numpy.array(space_group_symop.r_23, dtype=float)
    r_31 = numpy.array(space_group_symop.r_31, dtype=float)
    r_32 = numpy.array(space_group_symop.r_32, dtype=float)
    r_33 = numpy.array(space_group_symop.r_33, dtype=float)

    chi_21, chi_31, chi_32 = chi_12, chi_13, chi_23

    c11, r11 = numpy.meshgrid(chi_11, r_11, indexing="ij")
    c22, r22 = numpy.meshgrid(chi_22, r_22, indexing="ij")
    c33, r33 = numpy.meshgrid(chi_33, r_33, indexing="ij")
    c12, r12 = numpy.meshgrid(chi_12, r_12, indexing="ij")
    c13, r13 = numpy.meshgrid(chi_13, r_13, indexing="ij")
    c23, r23 = numpy.meshgrid(chi_23, r_23, indexing="ij")
    c21, r21 = numpy.meshgrid(chi_21, r_21, indexing="ij")
    c31, r31 = numpy.meshgrid(chi_31, r_31, indexing="ij")
    c32, r32 = numpy.meshgrid(chi_32, r_32, indexing="ij")

    rcrt_11, rcrt_12, rcrt_13, rcrt_21, rcrt_22, rcrt_23, rcrt_31, rcrt_32, \
        rcrt_33 = calc_mRmCmRT(
            (r11, r12, r13, r21, r22, r23, r31, r32, r33),
            (c11, c12, c13, c21, c22, c23, c31, c32, c33))

    # dimension (hkl, atoms, symmetry)
    n_a = numpy.newaxis
    fft_11 = ff[:, :, n_a] * rcrt_11[n_a, :, :]
    fft_12 = ff[:, :, n_a] * rcrt_12[n_a, :, :]
    fft_13 = ff[:, :, n_a] * rcrt_13[n_a, :, :]
    fft_21 = ff[:, :, n_a] * rcrt_21[n_a, :, :]
    fft_22 = ff[:, :, n_a] * rcrt_22[n_a, :, :]
    fft_23 = ff[:, :, n_a] * rcrt_23[n_a, :, :]
    fft_31 = ff[:, :, n_a] * rcrt_31[n_a, :, :]
    fft_32 = ff[:, :, n_a] * rcrt_32[n_a, :, :]
    fft_33 = ff[:, :, n_a] * rcrt_33[n_a, :, :]
    # ortogonalization should be done
    return fft_11, fft_12, fft_13, fft_21, fft_22, fft_23, fft_31, fft_32, \
        fft_33

def calc_moment_2d_by_susceptibility(r_ij, susc_i, m_norm_ij, h_loc):
    """Recalculate chi_i given according to symmetry elements for each point.

    chi_i  is given in reciprocal unit cell.

    After susceptibility is multiplied in magnetic field defined 

    r_ij:= r_11, r_12, r_13, r_21, r_22, r_23, r_31, r_32, r_33 
    susc_i:= chi_11, chi_22, chi_33, chi_12, chi_13, chi_23

    Matrix m_norm  used to recalculate coordinates  from direct space
    (a1/|a1|, a2/|a2|, a3/|a3|) to Cartesian one (x||a*, z||c).

    x_cart = m_norm * x_direct

    m_norm  = [[(1 - cos**2 alpha1 - cos**2 alpha2 - cos**2 alpha3 + \
                 2 cos alpha1 cos alpha2 cos alpha3)**0.5/sin(alpha1),  0,  0],
           [(cos alpha3 - cos alpha1 cos alpha2) / sin alpha1, sin alpha1,  0],
           [cos alpha2, cos alpha1,  1]]

    Matrix m_norm should be given as

        m_norm_ij = (_11, _12, _13, _21, _22, _23, _31, _32, _33) 

    Output
    ------
        moment_2d: [points, symmetry]
    """
    n_a = numpy.newaxis
    r_11, r_12, r_13, r_21, r_22, r_23, r_31, r_32, r_33 = r_ij
    chi_11, chi_22, chi_33, chi_12, chi_13, chi_23 = susc_i
    # [ind, symm]
    chi_2d_ij = calc_mRmCmRT(
        (r_11[n_a, :], r_12[n_a, :], r_13[n_a, :], r_21[n_a, :], r_22[n_a, :],
         r_23[n_a, :], r_31[n_a, :], r_32[n_a, :], r_33[n_a, :]),
        (chi_11[:, n_a], chi_12[:, n_a], chi_13[:, n_a], chi_12[:, n_a],
         chi_22[:, n_a], chi_23[:, n_a], chi_13[:, n_a], chi_23[:, n_a],
         chi_33[:, n_a]))
    chi_orto_ij = ortogonalize_matrix(chi_2d_ij, m_norm_ij)
    moment_2d = calc_product_matrix_vector(chi_orto_ij, h_loc)
    return moment_2d


def ortogonalize_matrix(m_ij, m_norm_ij):
    """Ortogonalize matrix.

    matrix m_ij is defined in coordinate system (a, b, c).
    It is given as tuple

    Matrix m_norm  used to recalculate coordinates  from direct space
    (a1/|a1|, a2/|a2|, a3/|a3|) to Cartesian one (x||a*, z||c).

    x_cart = m_norm * x_direct

    m_norm  = [[(1 - cos**2 alpha1 - cos**2 alpha2 - cos**2 alpha3 +  \
                 2 cos alpha1 cos alpha2 cos alpha3)**0.5/sin(alpha1),  0,  0],
           [(cos alpha3 - cos alpha1 cos alpha2) / sin alpha1, sin alpha1,  0],
           [cos alpha2, cos alpha1,  1]]

    Should be given as

        m_norm_ij = (_11, _12, _13, _21, _22, _23, _31, _32, _33)

    output matrix s_ij is defined in Cartezian coordinate system defined as
    x||a*, z||c, y= [z x] (right handed)
    """
    s_11, s_12, s_13, s_21, s_22, s_23, s_31, s_32, s_33 = calc_mRmCmRT(
        m_norm_ij, m_ij)
    return s_11, s_12, s_13, s_21, s_22, s_23, s_31, s_32, s_33


def calc_atoms_in_unit_cell(r_ij, b_i, fract_xyz, atom_label):
    """Calculate atoms in unit cell.

    Arguments
    ---------
        - r_ij = (r_11, r_12, r_13, r_21, r_22, r_23, r_31, r_32, r_33)
        - b_i = (b_1, b_2, b_3)
        - fract_xyz = (fract_x, fract_y, fract_z)
        - atom_label

    Output
    ------
        - fract_uc_x
        - fract_uc_y
        - fract_uc_z
        - label_uc
    """
    (r_11, r_12, r_13, r_21, r_22, r_23, r_31, r_32, r_33) = r_ij
    (b_1, b_2, b_3) = b_i
    atom_number = numpy.linspace(0, atom_label.size-1, atom_label.size)
    (fract_x, fract_y, fract_z) = fract_xyz
    na = numpy.newaxis
    f_x = numpy.mod(r_11[:, na]*fract_x[na, :] + r_12[:, na]*fract_y[na, :] +
                    r_13[:, na]*fract_z[na, :] + b_1[:, na], 1.)
    f_y = numpy.mod(r_21[:, na]*fract_x[na, :] + r_22[:, na]*fract_y[na, :] +
                    r_23[:, na]*fract_z[na, :] + b_2[:, na], 1.)
    f_z = numpy.mod(r_31[:, na]*fract_x[na, :] + r_32[:, na]*fract_y[na, :] +
                    r_33[:, na]*fract_z[na, :] + b_3[:, na], 1.)
    r_f_x = numpy.round(f_x, decimals=5)
    r_f_y = numpy.round(f_y, decimals=5)
    r_f_z = numpy.round(f_z, decimals=5)

    a_n = numpy.ones(r_f_x.shape, dtype=float) * atom_number[na, :]
    r_f_xyz = numpy.array([r_f_x.flatten(), r_f_y.flatten(), r_f_z.flatten(),
                           a_n.flatten()],
                          dtype=float)
    u_f_xyz = numpy.unique(r_f_xyz, axis=1)

    fract_uc_x, fract_uc_y, fract_uc_z = u_f_xyz[0], u_f_xyz[1], u_f_xyz[2]
    number_uc = numpy.round(u_f_xyz[3], decimals=0).astype(int)
    label_uc = atom_label[number_uc]
    ind_sort = numpy.argsort(number_uc)

    fract_uc_x_s = numpy.take_along_axis(fract_uc_x, ind_sort, 0)
    fract_uc_y_s = numpy.take_along_axis(fract_uc_y, ind_sort, 0)
    fract_uc_z_s = numpy.take_along_axis(fract_uc_z, ind_sort, 0)
    label_uc_s = numpy.take_along_axis(label_uc, ind_sort, 0)

    return fract_uc_x_s, fract_uc_y_s, fract_uc_z_s, label_uc_s



import os
import sys
if not(os.getcwd()  in sys.path):
    sys.path.append(os.getcwd())

from cryspy.D_functions_item_loop.function_1_report_magnetization_ellipsoid import \
        magnetization_ellipsoid_by_u_ij, AtomSiteSusceptibility, Cell


def test_uij():


    cell = Cell(
        length_a = 9.90710, length_b = 12.70190, length_c = 14.15560,
        angle_alpha = 90., angle_beta = 92.82, angle_gamma = 90.)
    
    a_s_s = AtomSiteSusceptibility(
        chi_11=0.350, chi_22=0.454, chi_33=2.620,
        chi_12=-0.299, chi_13=0.233, chi_23=1.066)
    
    u_ij = magnetization_ellipsoid_by_u_ij(cell, a_s_s)
    print(u_ij)
    assert True



import numpy
from typing import NoReturn
import scipy

from cryspy.A_functions_base.function_1_objects import \
    form_items_by_dictionary

from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN

from cryspy.C_item_loop_classes.cl_1_pd_meas import PdMeasL

na = numpy.newaxis


class TOFBackgroundPoint(ItemN):
    """Background point in powder 1d experiment.
    """
    ATTR_MANDATORY_NAMES = ("time", "intensity")
    ATTR_MANDATORY_TYPES = (float, float)
    ATTR_MANDATORY_CIF = ("time", "intensity")

    ATTR_OPTIONAL_NAMES = ()
    ATTR_OPTIONAL_TYPES = ()
    ATTR_OPTIONAL_CIF = ()

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ()
    ATTR_INT_PROTECTED_NAMES = ()

    # parameters considered are refined parameters
    ATTR_REF = ("intensity", )
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])
    ATTR_CONSTR_MARK = tuple([f"{_h:}_mark" for _h in ATTR_REF])

    # constraints on the parameters
    D_CONSTRAINTS = {}

    # default values for the parameters
    D_DEFAULT = {}
    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False
    for key in ATTR_CONSTR_MARK:
        D_DEFAULT[key] = ""

    PREFIX = "tof_backgroundpoint"

    def __init__(self, **kwargs) -> NoReturn:
        super(TOFBackgroundPoint, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {"time": 0}

        # defined for ani integer and float parameters
        D_MAX = {}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)


class TOFBackgroundPointL(LoopN):
    """Background points in powder 1d experiment.
    """
    ITEM_CLASS = TOFBackgroundPoint
    ATTR_INDEX = "time"

    def __init__(self, loop_name: str = None, **kwargs) -> NoReturn:
        super(TOFBackgroundPointL, self).__init__()
        self.__dict__["items"] = form_items_by_dictionary(
            self.ITEM_CLASS, kwargs)
        self.__dict__["loop_name"] = loop_name

    def interpolate_by_points(self, np_time):
        l_time = self.time
        l_intensity = self.intensity
        time_b = numpy.array(l_time, dtype=float)
        int_b = numpy.array(l_intensity, dtype=float)
        if len(l_time) == 0:
            int_1d = numpy.zeros(np_time.size, dtype=float)
        else:
            int_1d = numpy.interp(np_time, time_b, int_b)
        return int_1d

    def get_dictionary(self):
        res = {}
        res["background_time"] = numpy.array(self.time, dtype=float)
        res["background_intensity"] = numpy.array(self.intensity, dtype=float)
        res["flags_background_intensity"] = numpy.array(
            self.intensity_refinement, dtype=bool)
        return res



"""Functions for magnetic section.

Functions
---------
    - get_j0_j2_by_symbol
"""
from cryspy.A_functions_base.database import DATABASE
from cryspy.A_functions_base.charge_form_factor import get_atomic_symbol_ion_charge_isotope_number_by_ion_symbol



def get_j0_j2_by_symbol(symbol: str):
    """Get <j0>, <j2> parameters by symbol."""

    j0_a2, j0_A1, j0_B1, j0_b2 = 0., 0., 0., 0.
    j0_C1, j0_c2, j0_D = 0., 0., 0.
    j2_a2, j2_A1, j2_B1, j2_b2 = 0., 0., 0., 0.
    j2_C1, j2_c2, j2_D = 0., 0., 0.
    
    d_mff = DATABASE["Magnetic Form Factor, tabulated"]
    atom_name, ion_charge, isotope_number = get_atomic_symbol_ion_charge_isotope_number_by_ion_symbol(symbol)
    try:
        j0 = d_mff[("j0_AaBbCcD", atom_name, ion_charge)]
        j0_A1, j0_a2, j0_B1, j0_b2, j0_C1, j0_c2, j0_D = j0[0], j0[1], j0[2], j0[3], j0[4], j0[5], j0[6] 
    except KeyError:
        pass
    try:
        j2 = d_mff[("j2_AaBbCcD", atom_name, ion_charge)]
        j2_A1, j2_a2, j2_B1, j2_b2, j2_C1, j2_c2, j2_D = j2[0], j2[1], j2[2], j2[3], j2[4], j2[5], j2[6]
    except KeyError:
        pass
    return j0_A1, j0_a2, j0_B1, j0_b2, j0_C1, j0_c2, j0_D, j2_A1, j2_a2, \
        j2_B1, j2_b2, j2_C1, j2_c2, j2_D



"""Describe AtomSiteSusceptibility, AtomSiteSusceptibilityL."""
from typing import NoReturn
import math
import numpy

from cryspy.A_functions_base.function_1_algebra import calc_m_sigma
from cryspy.A_functions_base.function_1_atomic_vibrations import \
    vibration_constraints
from cryspy.A_functions_base.function_1_objects import \
    form_items_by_dictionary

import cryspy.A_functions_base.local_susceptibility as local_susceptibility

from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN

na = numpy.newaxis

# FIXME: temporary solution which is enough slow
def calc_independent_lines(m):
    n_x, n_y = m.shape
    l_ind = [0, ]
    for ind_x in range(1, n_x):
        det_m = numpy.linalg.det(m[l_ind+[ind_x, ], :][:, l_ind+[ind_x, ]])
        if numpy.isclose(det_m, 0.):
            pass
        else:
            l_ind.append(ind_x)
    return l_ind


def calc_constr_matrix(matrix):
    m_x, m_y = matrix.shape
    m_rank = numpy.linalg.matrix_rank(matrix)
    unity = numpy.diag(m_x*[1.])
    m_u = matrix - unity


    for i_x in range(m_x-1, 0, -1):
        coeff_1 = m_u[i_x, i_x]
        if not(numpy.isclose(coeff_1, 0., atol=1e-10)):
            line_1 = m_u[i_x, :]/coeff_1
            m_u[i_x, :] = line_1
            for i_y in range(i_x-1, -1, -1):
                coeff_2 = m_u[i_y, i_x]
                line_2 = m_u[i_y, :] - line_1*coeff_2
                m_u[i_y, :]  = line_2

    res = numpy.round(unity-m_u, 10)
    return res

class AtomSiteSusceptibility(ItemN):
    """Magnetic properties of the atom that occupy the atom site.

    Data items in the ATOM_SITE_MAGNETISM_ANISO category record details about
    magnetic properties of the atoms that occupy the atom sites.

    Attributes
    ----------
        -
    """

    ATTR_MANDATORY_NAMES = ("label", )
    ATTR_MANDATORY_TYPES = (str, )
    ATTR_MANDATORY_CIF = ("label", )

    ATTR_OPTIONAL_NAMES = (
        "chi_type", "chi_11", "chi_22", "chi_33", "chi_12",
        "chi_13", "chi_23", )
    ATTR_OPTIONAL_TYPES = (str, float, float, float, float, float, float, )
    ATTR_OPTIONAL_CIF = (
        "chi_type",  "chi_11", "chi_22", "chi_33", "chi_12",
        "chi_13", "chi_23", )

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ()
    ATTR_INT_PROTECTED_NAMES = ()

    # parameters considered are refined parameters
    ATTR_REF = ("chi_11", "chi_22", "chi_33", "chi_12", "chi_13", "chi_23",)
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])
    ATTR_CONSTR_MARK = tuple([f"{_h:}_mark" for _h in ATTR_REF])

    # formats if cif format
    D_FORMATS = {"chi_11": "{:.5f}", "chi_22": "{:.5f}", "chi_33": "{:.5f}",
                 "chi_12": "{:.5f}", "chi_13": "{:.5f}", "chi_23": "{:.5f}",}

    # constraints on the parameters
    D_CONSTRAINTS = {"chi_type": ["Ciso", "Cani"],
                     }

    # default values for the parameters
    D_DEFAULT = {}
    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False
    for key in ATTR_CONSTR_MARK:
        D_DEFAULT[key] = ""

    PREFIX = "atom_site_susceptibility"

    def __init__(self, **kwargs) -> NoReturn:
        super(AtomSiteSusceptibility, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {}

        # defined for ani integer and float parameters
        D_MAX = {}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)

    def apply_space_group_constraint(self, atom_site, space_group, cell):
        """
        Space group constraints.

        According to table 1 in Peterse, Palm, Acta Cryst.(1966), 20, 147
        """
        label_aniso = self.label
        label = atom_site.label
        index = label.index(label_aniso)

        item_as = atom_site.items[index]
        sc_chi = item_as.calc_sc_chi(space_group, cell)
        # it should be checked
        sc_chi = calc_constr_matrix(sc_chi)

        # l_numb = atom_site.calc_constr_number(space_group)

        flag_chi = self.is_attribute("chi_type")
        if flag_chi:
            flag_chi = self.chi_type.lower().startswith("cani")

        if flag_chi:
            self.__dict__["chi_11_constraint"] = False
            self.__dict__["chi_22_constraint"] = False
            self.__dict__["chi_33_constraint"] = False
            self.__dict__["chi_12_constraint"] = False
            self.__dict__["chi_13_constraint"] = False
            self.__dict__["chi_23_constraint"] = False

        if flag_chi:
            chi_i = numpy.array([
                self.chi_11, self.chi_22, self.chi_33,
                self.chi_12, self.chi_13, self.chi_23], dtype=float)

            chi_sigma_i = numpy.array([
                self.chi_11_sigma, self.chi_22_sigma, self.chi_33_sigma,
                self.chi_12_sigma, self.chi_13_sigma, self.chi_23_sigma], dtype=float)

            chi_ref_i = numpy.array([
                self.chi_11_refinement, self.chi_22_refinement, self.chi_33_refinement,
                self.chi_12_refinement, self.chi_13_refinement, self.chi_23_refinement], dtype=bool)

            chi_i_c = sc_chi.dot(chi_i)
            chi_sigma_i_c = numpy.sqrt(numpy.square(sc_chi).dot(numpy.square(chi_sigma_i)))
            l_ind_independent = calc_independent_lines(sc_chi)

            chi_con_i_c = numpy.ones(chi_i_c.shape, dtype=bool)
            chi_con_i_c[l_ind_independent] = False
            sc_chi_bool = numpy.logical_not(numpy.isclose(numpy.round(sc_chi, decimals=5), 0.))
            # chi_con_i_c = numpy.triu((sc_chi_bool[na, :, :] == sc_chi_bool[:, na, :]).all(axis=2), k=1).any(axis=0)
            chi_ref_i_c = sc_chi_bool.dot(chi_ref_i) * numpy.logical_not(chi_con_i_c)
            # # I am not quite sure. It is to fix two parameters among three.
            # chi_ref_i_c = numpy.logical_and(chi_ref_i_c, chi_ref_i)

            self.__dict__["chi_11"], self.__dict__["chi_22"], \
                self.__dict__["chi_33"], self.__dict__["chi_12"], \
                self.__dict__["chi_13"], self.__dict__["chi_23"] = \
                    chi_i_c[0], chi_i_c[1], chi_i_c[2], chi_i_c[3], chi_i_c[4], chi_i_c[5]

            self.__dict__["chi_11_sigma"], self.__dict__["chi_22_sigma"], \
                self.__dict__["chi_33_sigma"], self.__dict__["chi_12_sigma"], \
                self.__dict__["chi_13_sigma"], self.__dict__["chi_23_sigma"] =\
                chi_sigma_i_c[0], chi_sigma_i_c[1], chi_sigma_i_c[2], chi_sigma_i_c[3], \
                chi_sigma_i_c[4], chi_sigma_i_c[5]

            self.__dict__["chi_11_refinement"], \
                self.__dict__["chi_22_refinement"], \
                self.__dict__["chi_33_refinement"], \
                self.__dict__["chi_12_refinement"], \
                self.__dict__["chi_13_refinement"], \
                self.__dict__["chi_23_refinement"] = chi_ref_i_c[0], chi_ref_i_c[1], chi_ref_i_c[2],\
                    chi_ref_i_c[3], chi_ref_i_c[4], chi_ref_i_c[5]

            self.__dict__["chi_11_constraint"], \
                self.__dict__["chi_22_constraint"], \
                self.__dict__["chi_33_constraint"], \
                self.__dict__["chi_12_constraint"], \
                self.__dict__["chi_13_constraint"], \
                self.__dict__["chi_23_constraint"] = chi_con_i_c[0], chi_con_i_c[1], chi_con_i_c[2],\
                    chi_con_i_c[3], chi_con_i_c[4], chi_con_i_c[5]

    def apply_chi_iso_constraint(self, cell):
        """Isotropic constraint on susceptibility."""
        c_a = cell.cos_a
        s_ib = cell.sin_ib
        s_ig = cell.sin_ig
        c_ib = cell.cos_ib
        c_ig = cell.cos_ig
        # not sure, it is better to check
        if not(self.is_attribute("chi_type")):
            return
        chi_type = self.chi_type
        if chi_type.lower().startswith("ciso"):
            self.__dict__["chi_22"] = self.chi_11
            self.__dict__["chi_33"] = self.chi_11
            self.__dict__["chi_12"] = 0. # self.chi_11*c_ig
            self.__dict__["chi_13"] = 0. # self.chi_11*c_ib
            self.__dict__["chi_23"] = 0. # self.chi_11*(c_ib*c_ig-s_ib*s_ig*c_a)
            self.__dict__["chi_22_sigma"] = self.chi_11_sigma
            self.__dict__["chi_33_sigma"] = self.chi_11_sigma
            self.__dict__["chi_12_sigma"] = 0. # self.chi_11_sigma * c_ig
            self.__dict__["chi_13_sigma"] = 0. # self.chi_11_sigma * c_ib
            self.__dict__["chi_23_sigma"] = 0. # self.chi_11_sigma * (c_ib*c_ig-s_ib*s_ig*c_a)
            self.__dict__["chi_22_refinement"] = False
            self.__dict__["chi_33_refinement"] = False
            self.__dict__["chi_12_refinement"] = False
            self.__dict__["chi_13_refinement"] = False
            self.__dict__["chi_23_refinement"] = False
            self.__dict__["chi_22_constraint"] = True
            self.__dict__["chi_33_constraint"] = True
            self.__dict__["chi_12_constraint"] = True
            self.__dict__["chi_13_constraint"] = True
            self.__dict__["chi_23_constraint"] = True

    def calc_main_axes_of_magnetization_ellipsoid(self, cell):
        """Susceptibility along the main axes of magnetization ellipsoid.

        Arguments
        ---------
            - cell

        Output
        ------
            - moments is main axes of ellipsoid in mu_B/T
            - moments_sigma is sigmas for main axes of ellipsoid
            - rot_matrix is directions for moments
                for moments[0] direction is rot_matrix[:, 0]
                for moments[1] direction is rot_matrix[:, 1]
                for moments[2] direction is rot_matrix[:, 2]

        The main axes are given in Cartezian coordinate system (x||a*, z||c).
        """

        ucp = numpy.array([
            cell.length_a, cell.length_b, cell.length_c,
            cell.angle_alpha*numpy.pi/180., cell.angle_beta*numpy.pi/180.,
            cell.angle_gamma*numpy.pi/180.], dtype=float)
        flag_ucp = False

        q_rn = numpy.array([
            self.chi_11, self.chi_22, self.chi_33,
            self.chi_12, self.chi_13, self.chi_23], dtype=float)
        flag_q_rn = True

        # FIXME: add errorbars calculations
        q_rn_sigma = numpy.array([
            self.chi_11_sigma, self.chi_22_sigma, self.chi_33_sigma,
            self.chi_12_sigma, self.chi_13_sigma, self.chi_23_sigma], dtype=float)

        moments, eig_fields, eig_moments = local_susceptibility.calc_magnetization_ellipsoid_axes(
            q_rn, ucp)

        moments_sigma = numpy.zeros_like(moments)
        return moments, moments_sigma, eig_fields, eig_moments

    def get_flags_susceptibility(self):
        res = numpy.array([
            self.chi_11_refinement, self.chi_22_refinement, self.chi_33_refinement,
            self.chi_12_refinement, self.chi_13_refinement, self.chi_23_refinement], dtype=bool)
        return res



class AtomSiteSusceptibilityL(LoopN):
    """Magnetic properties of the atoms that occupy the atom sites.

    Methods
    -------
        - apply_space_group_constraint
        - apply_chi_iso_constraint
    """

    ITEM_CLASS = AtomSiteSusceptibility
    ATTR_INDEX = "label"

    def __init__(self, loop_name: str = None, **kwargs) -> NoReturn:
        super(AtomSiteSusceptibilityL, self).__init__()
        self.__dict__["items"] = form_items_by_dictionary(self.ITEM_CLASS, kwargs)
        self.__dict__["loop_name"] = loop_name

    def apply_space_group_constraint(self, atom_site, space_group, cell):
        """Apply space group constraint."""
        for item in self.items:
            item.apply_space_group_constraint(atom_site, space_group, cell)

    def apply_chi_iso_constraint(self, cell):
        """Apply isotropic constraint on susceptibility."""
        for item in self.items:
            item.apply_chi_iso_constraint(cell)


    def calc_main_axes_of_magnetization_ellipsoid(self, cell):
        """Susceptibility along the main axes of magnetization ellipsoid.

        Arguments
        ---------
            - cell

        Output
        ------
            - l_moments is main axes of ellipsoid in mu_B/T for each atom
            - l_moments_sigma is sigmas for main axes of ellipsoid for each
              atom
            - l_rot_matrix is directions for moments
                for moments[0] direction is rot_matrix[:, 0]
                for moments[1] direction is rot_matrix[:, 1]
                for moments[2] direction is rot_matrix[:, 2]

        The main axes are given in Cartezian coordinate system (x||a*, z||c).
        """
        l_moments, l_moments_sigma, l_eig_fields, l_eig_moments = [], [], [], []
        for item in self.items:
            moments, moments_sigma, eig_fields, eig_moments = \
                item.calc_main_axes_of_magnetization_ellipsoid(cell)
            l_moments.append(moments)
            l_moments_sigma.append(moments_sigma)
            l_eig_fields.append(eig_fields)
            l_eig_moments.append(eig_moments)
        return l_moments, l_moments_sigma, l_eig_fields, l_eig_moments


    def get_flags_susceptibility(self):
        flags_susceptibility = numpy.stack([item.get_flags_susceptibility() for item in self.items], axis=0).transpose()
        return flags_susceptibility




"""Module realized some operations with python objects

"""
from types import FunctionType


def get_functions_of_objet(obj):
    """Get list of object function.
    """
    ls_out = []
    l_method = [_1 for _1, _2 in type(obj).__dict__.items()
                if ((type(_2) == FunctionType) &
                    (not(_1.startswith("_"))))]
    for method in l_method:
        func = getattr(obj, method)
        l_param = [_ for _ in
                   func.__code__.co_varnames[:func.__code__.co_argcount]
                   if _ != "self"]
        s_par = ""
        if len(l_param) > 0:
            s_par = ", ".join(l_param)
        s_val = f"{method:}({s_par:})"
        ls_out.append(s_val)
    return sorted(ls_out)


def variable_name_to_string(variable_name: tuple, flag_html: bool = True):
    """Give string presentation of variable name.
    
    Variable name: 
        ((name, val), (name, val), ...)
    """
    ls_out = []
    for name_val in variable_name[1:]:
        name, val = name_val
        if ((val is None) | (val == "")):
            ls_out.append(f"{name_val[0]:}")
        elif isinstance(name_val[1], int):
            ls_out.append(f"{name_val[0]:}[{name_val[1]:}]")
        else:
            ls_out.append(f"{name_val[0]:}_{name_val[1]:}")
    if flag_html:
        ls_out[-1] = "<b>"+ls_out[-1]+"</b>"
    return '.'+'.'.join(ls_out)


def change_variable_name(variable_name: tuple, ending: str = ""):
    """Change the ending of variable name.
    
    Output:
        modified variable name
    """
    if ending == "":
        res = variable_name
    elif len(variable_name[-1]) == 1:
        res = variable_name[:-1] + ((f"{variable_name[-1][0]:}_{ending:}", ), )
    else:
        res = variable_name[:-1] + ((f"{variable_name[-1][0]:}_{ending:}", 
                                     variable_name[-1][1]), )
    return res


def get_table_html_for_variables(obj):
    """Get html table for variables of object
    """
    ls_html=["<table border='1'>"]
    variable_names = obj.get_variable_names()
    ls_variables = []
    if len(variable_names) != 0:
        ls_variables.append("<th>Variable</th><th>Value</th><th>Sigma</th>")
        ls_variables.extend([f"<td>{variable_name_to_string(var_name):} </td> \
<td>{obj.get_variable_by_name(var_name):}</td>\
<td>{obj.get_variable_by_name(change_variable_name(var_name, 'sigma')):}</td>"
for var_name in variable_names])

    for variables in ls_variables:
        ls_html.append(f"<tr>{variables}</tr>")
    ls_html.append("</table>")

    return " ".join(ls_html)


def form_items_by_dictionary(item_class, d_items):
    """
    Form items for loop when variables are given by dictionary
    """
    items = []
    if "items" in d_items.items():
        items  = d_items["items"]
    elif len(d_items.items()) > 0:
        i=0
        for hh in zip(*d_items.items()):
            if i==0:
                keys = hh
                i += 1
            else:
                vals = hh
        
        for val in zip(*vals):
            d_item = {}
            for k,v in zip(keys, val):
                d_item[k]=v
            item = item_class(**d_item)
            items.append(item)
        
    return items


"""Calculation of flip ratio.

Functions
---------
    - calc_flip_ratio
    - calc_intensity_plus_down
    - calc_fm_perp_loc
    - calc_fm_perp_for_fm_loc
    - calc_fm_loc
    - calc_e_up_loc
    - calc_rotated_ub

"""
import numpy

from cryspy.A_functions_base.function_3_extinction import calc_extinction_2

from cryspy.C_item_loop_classes.cl_1_cell import Cell
from cryspy.C_item_loop_classes.cl_1_extinction import Extinction
from cryspy.C_item_loop_classes.cl_1_setup import Setup
from cryspy.C_item_loop_classes.cl_1_diffrn_radiation import DiffrnRadiation
from cryspy.C_item_loop_classes.cl_2_diffrn_orient_matrix import \
    DiffrnOrientMatrix


def calc_flip_ratio(
        sthovl: numpy.ndarray, wavelength: float, field_norm: float,
        u_matrix: numpy.ndarray, polarization: float,
        flipper_efficiency: float, volume_unit_cell: float,
        model_extinction: str, radius: float, mosaicity: float,
        f_nucl: numpy.array, fm_perp_loc: tuple,
        ratio_lambdaover2: float = None, f_nucl_2hkl: numpy.array = None,
        fm_perp_loc_2hkl: tuple = None):
    """Calculate flip ratio."""
    dder = {}
    iint_u, iint_d, dder_iint = calc_intensity_plus_down(
        sthovl, wavelength, field_norm, u_matrix, polarization,
        flipper_efficiency, f_nucl, fm_perp_loc, volume_unit_cell,
        model_extinction, radius, mosaicity)

    if ratio_lambdaover2 is not None:
        sthovl_2hkl = 2.0 * sthovl
        wavelength_2hkl = 0.5*wavelength
        polarization_2hkl = 0.

        iint_u_2hkl, iint_d_2hkl, dder_iint_2hkl = calc_intensity_plus_down(
            sthovl_2hkl, wavelength_2hkl, field_norm, u_matrix,
            polarization_2hkl, flipper_efficiency, f_nucl_2hkl,
            fm_perp_loc_2hkl, volume_unit_cell, model_extinction, radius,
            mosaicity)

        flip_ratio = (iint_u+ratio_lambdaover2*iint_u_2hkl) / \
            (iint_d+ratio_lambdaover2*iint_d_2hkl)
    else:
        flip_ratio = iint_u / iint_d

    return flip_ratio, dder


def calc_intensity_plus_down(
        sthovl: numpy.ndarray, wavelength: float, field_norm: float,
        u_matrix: numpy.ndarray, polarization: float,
        flipper_efficiency: float, f_nucl: numpy.array,
        fm_perp_loc: tuple, volume_unit_cell: float,
        model_extinction: str, radius: float, mosaicity: float):
    """Calculate integrated intensity up and down."""
    dder = {}

    p_u = polarization
    p_d = (2.*flipper_efficiency-1.)*polarization

    phi_d, chi_d, omega_d = 0., 0., 0.
    e_up_loc = calc_e_up_loc(phi_d, chi_d, omega_d, u_matrix)

    mag_p_1, mag_p_2, mag_p_3 = fm_perp_loc

    mag_p_sq = abs(mag_p_1*mag_p_1.conjugate() +
                   mag_p_2*mag_p_2.conjugate() +
                   mag_p_3*mag_p_3.conjugate())

    mag_p_e_u = mag_p_1*e_up_loc[0]+mag_p_2*e_up_loc[1]+mag_p_3*e_up_loc[2]

    f_nucl_sq = abs(f_nucl)**2
    mag_p_e_u_sq = abs(mag_p_e_u*mag_p_e_u.conjugate())
    fnp = (mag_p_e_u*f_nucl.conjugate()+mag_p_e_u.conjugate()*f_nucl).real
    # fp_sq = f_nucl_sq + mag_p_sq + fnp
    # fm_sq = f_nucl_sq + mag_p_sq - fnp
    fp_sq = f_nucl_sq + mag_p_e_u_sq + fnp
    fm_sq = f_nucl_sq + mag_p_e_u_sq - fnp
    fpm_sq = mag_p_sq - mag_p_e_u_sq

    l_model_extinction = ["gauss", "lorentz"]
    if model_extinction.lower() in l_model_extinction:
        yp, dder_yp = calc_extinction_2(
            radius, mosaicity, model_extinction, fp_sq, volume_unit_cell,
            sthovl, wavelength)

        ym, dder_ym = calc_extinction_2(
            radius, mosaicity, model_extinction, fm_sq, volume_unit_cell,
            sthovl, wavelength)

        ypm, dder_ypm = calc_extinction_2(
            radius, mosaicity, model_extinction, fpm_sq, volume_unit_cell,
            sthovl, wavelength)

    else:
        yp = 1. + 0.*f_nucl_sq
        ym = yp
        ypm = yp

    iint_u = 0.5*(1.+p_u)*yp*fp_sq + 0.5*(1.-p_u)*ym*fm_sq + ypm*fpm_sq
    iint_d = 0.5*(1.-p_d)*yp*fp_sq + 0.5*(1.+p_d)*ym*fm_sq + ypm*fpm_sq
    # print("iint_u_1\n", iint_u, "\n")
    # print("iint_d_1\n", iint_d, "\n")

    # pppl = 0.5*((1+p_u)*yp+(1-p_u)*ym)
    # ppmin = 0.5*((1-p_d)*yp+(1+p_d)*ym)
    # pmpl = 0.5*((1+p_u)*yp-(1-p_u)*ym)
    # pmmin = 0.5*((1-p_d)*yp-(1+p_d)*ym)

    # # integral intensities and flipping ratios
    # iint_u = (f_nucl_sq+mag_p_e_u_sq)*pppl + pmpl*fnp + ypm*fpm_sq
    # iint_d = (f_nucl_sq+mag_p_e_u_sq)*ppmin + pmmin*fnp + ypm*fpm_sq
    # print("iint_u_2\n", iint_u, "\n")
    # print("iint_d_2\n", iint_d, "\n")

    return iint_u, iint_d, dder


def calc_fm_perp_loc(e_up_loc: numpy.ndarray, field_norm: float,
                     k_loc_i: tuple, sft_ij: tuple, sftm_ij: tuple):
    """Calculate the magnetic structure factor perpendicular to hkl.

    Cartezian coordinate system is defined such way that
    (x||a*), (z||c).
    """

    fm_loc = calc_fm_loc(e_up_loc, sft_ij, sftm_ij, field_norm)
    fm_perp = calc_fm_perp_for_fm_loc(k_loc_i, fm_loc)
    mag_p_1, mag_p_2, mag_p_3 = fm_perp

    return mag_p_1, mag_p_2, mag_p_3


def calc_fm_perp_for_fm_loc(k_loc_i, fm_loc):
    """Calculate perpendicular component of fm to scattering vector."""
    k_1, k_2, k_3 = k_loc_i[0], k_loc_i[1], k_loc_i[2]
    mag_1, mag_2, mag_3 = fm_loc[0], fm_loc[1], fm_loc[2]
    mag_p_1 = (k_3*mag_1 - k_1*mag_3)*k_3 - (k_1*mag_2 - k_2*mag_1)*k_2
    mag_p_2 = (k_1*mag_2 - k_2*mag_1)*k_1 - (k_2*mag_3 - k_3*mag_2)*k_3
    mag_p_3 = (k_2*mag_3 - k_3*mag_2)*k_2 - (k_3*mag_1 - k_1*mag_3)*k_1
    return mag_p_1, mag_p_2, mag_p_3


def calc_fm_loc(e_up_loc: numpy.ndarray, sft_ij: tuple, sftm_ij: tuple,
                field_norm: float):
    """Calculate the magnetic structure factor for hkl.

    Output
    ------
        x, y and z coordinates of FM_perp

    Cartezian coordinate system is defined such way that
    (x||a*), (z||c).
    """
    sft_11, sft_12, sft_13, sft_21, sft_22, sft_23, sft_31, sft_32, sft_33 = \
        sft_ij
    sftm_11, sftm_12, sftm_13, sftm_21, sftm_22, sftm_23, sftm_31, sftm_32, \
        sftm_33 = sftm_ij

    field_loc = field_norm * e_up_loc

    # not sure about e_u_loc at field < 0
    mag_1 = sft_11*field_loc[0] + sft_12*field_loc[1] + \
        sft_13*field_loc[2] + sftm_11*e_up_loc[0] + sftm_12*e_up_loc[1] + \
        sftm_13*e_up_loc[2]
    mag_2 = sft_21*field_loc[0] + sft_22*field_loc[1] + \
        sft_23*field_loc[2] + sftm_21*e_up_loc[0] + sftm_22*e_up_loc[1] + \
        sftm_23*e_up_loc[2]
    mag_3 = sft_31*field_loc[0] + sft_32*field_loc[1] + \
        sft_33*field_loc[2] + sftm_31*e_up_loc[0] + sftm_32*e_up_loc[1] + \
        sftm_33*e_up_loc[2]

    return mag_1, mag_2, mag_3


def calc_e_up_loc(phi_d: float, chi_d: float, omega_d: float,
                  u_matrix: numpy.ndarray):
    """Calculate vertical direction in local Cartezian coordinate system.

    Local Cartezian coordinate system is defined such way that
    (x||a*), (z||c).
    """
    m_u_rot = calc_rotated_ub(phi_d, chi_d, omega_d, u_matrix)
    e_u_vec = numpy.array([0., 0., 1.], dtype=float)
    e_u_loc = numpy.matmul(m_u_rot.transpose(), e_u_vec)
    return e_u_loc


def calc_rotated_ub(phi_d: float, chi_d: float, omega_d: float,
                    u_matrix: numpy.ndarray):
    """Rotated U matrix by three angles."""
    m_phi_d = numpy.array([[numpy.cos(phi_d), numpy.sin(phi_d), 0.],
                           [-numpy.sin(phi_d), numpy.cos(phi_d), 0.],
                           [0., 0., 1.]], dtype=float)

    m_omega_d = numpy.array([[numpy.cos(omega_d), numpy.sin(omega_d), 0.],
                             [-numpy.sin(omega_d), numpy.cos(omega_d), 0.],
                             [0., 0., 1.]], dtype=float)

    m_chi_d = numpy.array([[numpy.cos(chi_d), 0., numpy.sin(chi_d)],
                           [0., 1., 0.],
                           [-numpy.sin(chi_d), 0., numpy.cos(chi_d)]],
                          dtype=float)

    m_u_rot = numpy.matmul(m_omega_d,
                           numpy.matmul(m_chi_d,
                                        numpy.matmul(m_phi_d, u_matrix)))
    return m_u_rot



from typing import List, Union
import numpy
import scipy
import scipy.optimize

from warnings import warn


from cryspy.A_functions_base.function_1_inversed_hessian import \
    estimate_inversed_hessian_matrix
from cryspy.A_functions_base.function_1_error_simplex import \
    error_estimation_simplex

from cryspy.B_parent_classes.cl_3_data import DataN
from cryspy.B_parent_classes.cl_4_global import GlobalN

from cryspy.C_item_loop_classes.cl_1_inversed_hessian import InversedHessian, inversed_hessian_to_correlation

from cryspy.E_data_classes.cl_1_crystal import Crystal
# from cryspy.E_data_classes.cl_1_mag_crystal import MagCrystal
from cryspy.E_data_classes.cl_2_diffrn import Diffrn
from cryspy.E_data_classes.cl_2_pd import Pd
from cryspy.E_data_classes.cl_2_pd2d import Pd2d
from cryspy.E_data_classes.cl_2_tof import TOF

from cryspy.procedure_rhochi.rhochi_by_dictionary import \
    rhochi_lsq_by_dictionary, rhochi_rietveld_refinement_by_dictionary,\
    rhochi_calc_chi_sq_by_dictionary

import cryspy

na = numpy.newaxis

def rhochi_check_items(cryspy_object: cryspy.GlobalN):
    """Check items for RhoChi refinement procedure.

    Output is list of experiments and list of crystals taken from cryspy_object
    """
    f_crystal, f_experiment = False, False

    if not(isinstance(cryspy_object, cryspy.GlobalN)):
        raise AttributeError("Incorrect type of object")

    cl_crysts = (cryspy.Crystal, ) #cryspy.MagCrystal
    cl_exps = (cryspy.Diffrn, cryspy.Pd, cryspy.Pd2d, cryspy.TOF)
    for item in cryspy_object.items:

        if isinstance(item, cl_crysts):
            f_crystal = True

        if isinstance(item, cl_exps):
            f_experiment = True

        if f_crystal & f_experiment:
            break

    if not(f_crystal):
        raise AttributeError("Crystal is not defined")

    if not(f_experiment):
        raise AttributeError("Experiment is not defined")



def rhochi_rietveld_refinement(cryspy_object: cryspy.GlobalN) -> dict:
    """Run refinement by RhoChi procedure.
    """
    # check object
    rhochi_check_items(cryspy_object)

    method = "BFGS"

    dict_out = rhochi_rietveld_refinement_with_parameters(
        cryspy_object,
        optimization_method=method) 

    return dict_out


def rhochi_rietveld_refinement_with_parameters(
        cryspy_object: cryspy.GlobalN,
        optimization_method: str = "BFGS") -> dict:
    """Run refinement by RhoChi procedure with non-default parameters.
    """

    # check object
    rhochi_check_items(cryspy_object)

    obj_dict = cryspy_object.get_dictionary()
    flag_scipy_refinements = True
    if flag_scipy_refinements:
        
        DICT_PARAMS["previous_arg"] = ()
        DICT_PARAMS["iteration"] = 0
        chi_sq, parameter_name, dict_in_out, res = rhochi_rietveld_refinement_by_dictionary(
            obj_dict, method=optimization_method, callback=_f_callback)
        dict_out = {"chi_sq": chi_sq, "parameter_name": parameter_name}
        if "hess_inv" in res.keys():
            hess_inv = res["hess_inv"]
            sigma_p = numpy.sqrt(numpy.abs(numpy.diag(hess_inv)))
            correlation_matrix = hess_inv/(sigma_p[:, na]*sigma_p[na, :])
            dict_out["correlation_matrix"] = correlation_matrix

            l_label = [hh[-1][0] for hh in parameter_name]
            if len(l_label) == hess_inv.shape[0]:
                inv_hessian = InversedHessian()
                inv_hessian.set_labels(l_label)
                inv_hessian.set_inversed_hessian(hess_inv)
                inv_hessian.form_inversed_hessian()
                inv_hessian.form_object()
                cryspy_object.add_items([inv_hessian, ])
        else:
            sigma_p = numpy.zeros((len(parameter_name),), dtype=float)
    else:
        chi_sq, delta_p, parameter_name, der_chi_sq, dder_chi_sq, dict_in_out = rhochi_lsq_by_dictionary(obj_dict)
        hessian = numpy.linalg.inv(dder_chi_sq)
        sigma_p = numpy.sqrt(numpy.diag(hessian))
        correlation_matrix = hessian/(sigma_p[:, na]*sigma_p[na, :])
        sigma_p[numpy.isnan(sigma_p)] = 0.
        correlation_matrix[numpy.isnan(correlation_matrix)] = 0.
        dict_out = {"chi_sq": chi_sq, "parameter_name": parameter_name,"der_chi_sq":der_chi_sq,
            "dder_chi_sq": dder_chi_sq, "correlation_matrix": correlation_matrix}
    cryspy_object.take_parameters_from_dictionary(obj_dict, l_parameter_name = parameter_name, l_sigma=sigma_p)
    cryspy_object.take_parameters_from_dictionary(dict_in_out, l_parameter_name = None, l_sigma=None)
    
    var_names = cryspy_object.get_variable_names()
    if len(var_names) > 0:
        print("Optimal parameters:")
    for name in var_names:
        value = cryspy_object.get_variable_by_name(name)
        print(f" - {name[-1][0]:} {value:.5f}")
    
    print("Errors are estimating by numerical derivatives.")
    rhochi_inversed_hessian(cryspy_object)
    return dict_out


def rhochi_no_refinement(cryspy_object: cryspy.GlobalN) -> dict:
    """Run calculations by RhoChi procedure.
    """
    flag_calc_analytical_derivatives = False
    flag_use_precalculated_data = False

    obj_dict = cryspy_object.get_dictionary()
    dict_in_out = {}

    chi_sq, n_point, der_chi_sq, dder_chi_sq, parameter_names = rhochi_calc_chi_sq_by_dictionary(
        obj_dict,
        dict_in_out=dict_in_out,
        flag_use_precalculated_data=flag_use_precalculated_data, flag_calc_analytical_derivatives=flag_calc_analytical_derivatives)
    
    cryspy_object.take_parameters_from_dictionary(obj_dict, l_parameter_name = None, l_sigma=None)
    cryspy_object.take_parameters_from_dictionary(dict_in_out, l_parameter_name = None, l_sigma=None)
    
    dict_out = {"chi_sq": chi_sq, "n_point": n_point}
    return dict_out

DICT_PARAMS = {"previous_arg": (), "iteration": 0}

def _f_callback(*arg, d_info: dict = None) -> bool:
    flag_out = False
    res_x = arg[0]
    if len(DICT_PARAMS["previous_arg"]) != len(res_x):
        DICT_PARAMS["previous_arg"] = res_x
    else:
        DICT_PARAMS["iteration"] += 1
        diff = numpy.array(res_x, dtype=float) - numpy.array(DICT_PARAMS["previous_arg"], dtype=float)
        shift = numpy.sqrt(numpy.square(diff).sum()) * 100
        print(f"Average shift of parameters is {shift:.5f} ({DICT_PARAMS['iteration']:}).                ", end="\r")
        DICT_PARAMS["previous_arg"] = res_x
    return flag_out
    # ls_out = ["{:12.5f}".format(_1) for _1 in res_x]
    # print(" ".join(ls_out), end="\r")
    # return flag_out


def rhochi_inversed_hessian(global_object: GlobalN):
    """Estimate inversed Hessian matrix."""
    if global_object.is_attribute("inversed_hessian"):
        global_object.items.remove(global_object.inversed_hessian)

    global_dict = global_object.get_dictionary()

    flag_calc_analytical_derivatives = False
    flag_use_precalculated_data = False

    obj_dict = global_object.get_dictionary()
    dict_in_out = {}

    chi_sq, n_point, der_chi_sq, dder_chi_sq, parameter_names = rhochi_calc_chi_sq_by_dictionary(
        obj_dict,
        dict_in_out=dict_in_out,
        flag_use_precalculated_data=flag_use_precalculated_data, flag_calc_analytical_derivatives=flag_calc_analytical_derivatives)
    
    param_0 = [global_dict[way[0]][way[1]][way[2]] for way  in parameter_names]

    flag_use_precalculated_data = True
    def tempfunc(l_param):
        for way, param in zip(parameter_names, l_param):
            global_dict[way[0]][way[1]][way[2]] = param
        chi_sq = rhochi_calc_chi_sq_by_dictionary(
            global_dict,
            dict_in_out=dict_in_out,
            flag_use_precalculated_data=flag_use_precalculated_data,
            flag_calc_analytical_derivatives=flag_calc_analytical_derivatives)[0]
        return chi_sq

    hess_inv, np_first_der = estimate_inversed_hessian_matrix(tempfunc, param_0)
    if (numpy.all(hess_inv == numpy.zeros_like(hess_inv)) or (hess_inv is None)):
        return 
    corr_matrix, sigmas = inversed_hessian_to_correlation(hess_inv) 
    global_object.take_parameters_from_dictionary(
        global_dict, l_parameter_name = parameter_names, l_sigma=sigmas)
    
    l_label = []
    for way in parameter_names:
        way_1, way_2, way_3 = way[0], way[1], way[2]
        l_h = way_1.split("_")
        if len(l_h) > 1:
            s_1 = "_".join(l_h[1:])
        else:
            s_1 = l_h[0]

        l_h = way_2.split("_")
        if len(l_h) > 1:
            s_2 = "_".join(l_h[1:])
        else:
            s_2 = l_h[0]

        s_3 = str(way_3).replace(" ", "")

        l_label.append(f"{s_1:},{s_2:},{s_3:}")


    # l_label = [f"{str(var_name).replace(' ',''):}" for i_param, var_name in enumerate(parameter_names)]

    inv_hessian = InversedHessian()
    inv_hessian.set_labels(l_label)
    inv_hessian.set_inversed_hessian(hess_inv)
    inv_hessian.form_inversed_hessian()
    inv_hessian.form_object()

    global_object.add_items([inv_hessian, ])

    return inv_hessian



"""
Module gives correction on extinction in case of isotropical case.
"""

import numpy

from .function_2_crystallography_base import \
    calc_volume_uc_by_abc_cosines,\
    calc_sthovl_by_hkl_abc_cosines

def calc_extinction(radius:float, mosaicity:float, model:str,
                    a:float, b:float, c:float, alpha:float, beta:float, gamma:float,
                    h:float, k:float, l:float, 
                    f_sq:float, wavelength:float, flag_derivative_f_sq=False):
    """
    Isotropical extinction coorection y:

    $$
    |F|^2_{\text{corrected}} = y \cdot |F|^2
    $$

    radius primary extinction ???
    mosaisity secondary extinction
    model= "gauss" or "lorentz"
    a,b,c,alpha,beta,gamma are unit cell parameters (in angstrem and radians)
    h, k, l are Miller indices
    f_sq is square of structure factor (in 10-12cm)
    wavelength is neutron wavelength in angstrems
    flag_derivative_radius
    flag_derivative_mosaicity
    flag_derivative_a
    flag_derivative_b
    flag_derivative_c
    flag_derivative_alpha
    flag_derivative_beta
    flag_derivative_gamma
    flag_derivative_f_sq
    flag_derivative_wavelength
    """
    r = float(radius)
    g = float(mosaicity)
    g_sq = numpy.square(g)

    kk = 1.
    c_a, c_b, c_g = numpy.cos(alpha), numpy.cos(beta), numpy.cos(gamma)
    
    volume_unit_cell = calc_volume_uc_by_abc_cosines(a, b, c, c_a, c_b, c_g)
    
    sthovl = calc_sthovl_by_hkl_abc_cosines(h, k, l, a, b, c, c_a, c_b, c_g)

    yext, dder  = calc_extinction_2(radius, mosaicity, model,
                                f_sq, volume_unit_cell, sthovl, wavelength) 
    return yext, dder 
    

def calc_extinction_2(radius:float, mosaicity:float, model:str,
                    f_sq:float, volume_unit_cell:float, sthovl:float, 
                    wavelength:float):
    """
    radius in micro meters
    g in minutes
    f_sq in 10**-12cm
    volume_unit_cell in angstrem**3
    wavelength in angstrem
    sthovl in angstrem**-1
    """
    r = radius
    g = mosaicity*numpy.pi/(180*60) #transfer minutes to radians
    g_sq = numpy.square(g)
    kk = 1.
    volume_sq = numpy.square(volume_unit_cell)
    stheta = sthovl * wavelength

    s2theta = 2.*stheta*numpy.sqrt(1.-numpy.square(stheta))
    c2theta = 1.-2.*numpy.square(stheta)
    c2theta_sq = numpy.square(c2theta)

    wavelength_cube = wavelength**3
    
    q = (f_sq*kk/volume_sq)*(wavelength_cube)*1./s2theta 
    delta_f_sq_q = (kk/volume_sq)*(wavelength_cube)*1./s2theta 

    t = 1.5*r
    alpha = 1.5*r*s2theta*1./wavelength
    x = 2./3*q*alpha*t

    x_sq = numpy.square(x)
    delta_f_sq_x = 2./3*delta_f_sq_q*alpha*t

    A = 0.20 + 0.45 * c2theta
    B = 0.22 - 0.12 * numpy.square(0.5-c2theta)
    yp = 1./(numpy.sqrt(1.+2.*x+(A*x_sq)*1./(1.+B*x)))
    yp_cube = yp**3
    delta_f_sq_yp = -0.5*(yp_cube)*(2.*delta_f_sq_x+
                      2.*A*delta_f_sq_x*x*1./(1.+B*x)-
                      B*delta_f_sq_x*(A*x_sq)*1./(numpy.square(1.+B*x))) 

    ag = 0.*sthovl 
    al = 0.*sthovl

    flag = alpha != 0.
    ag[flag] = alpha[flag]*g*(g_sq+0.5*alpha[flag]**2)**(-0.5)
    al[flag] = alpha[flag]*g*1./(g+alpha[flag]*2./3.)

    if model == "gauss":
        xs = 2./3.*q*ag*t
        xs_sq = numpy.square(xs)
        delta_f_sq_xs = 2./3.*delta_f_sq_q*ag*t
        A = 0.58 + 0.48 * c2theta + 0.24 * c2theta_sq
        B = 0.02 - 0.025 * c2theta
        #print("A, B", A, B)
        ys = 1./numpy.sqrt(1+2.12*xs+(A*xs_sq)*1./(1+B*xs))
        ys_cube = ys**3
        delta_f_sq_ys = -0.5*ys_cube*\
                        (2.12*delta_f_sq_xs+\
                        2.*(A*xs*delta_f_sq_xs)*1./(1+B*xs)-\
                        B*delta_f_sq_xs*(A*xs_sq)*1./(numpy.square(1+B*xs)))
    elif model == "lorentz":
        xs = 2./3.*q*al*t
        xs_sq = numpy.square(xs)
        delta_f_sq_xs = 2./3.*delta_f_sq_q*al*t

        A = 0.025 + 0.285 * c2theta
        B = -0.45 * c2theta
        flag = c2theta>0
        B[flag] = 0.15 - 0.2 * numpy.square(0.75-c2theta[flag])
        ys = 1./numpy.sqrt(1.+2.*xs+(A*xs_sq)*1./(1+B*xs))
        ys_cube = ys**3
        delta_f_sq_ys = -0.5*ys_cube*\
                        (2.*delta_f_sq_xs+\
                        2.*(A*xs*delta_f_sq_xs)*1./(1+B*xs)-\
                        B*delta_f_sq_xs*(A*xs_sq)*1./(numpy.square(1.+B*xs)))
    else:
        ys = 1.
        delta_f_sq_ys = 0.
    #print("ys", ys)
    yext = yp * ys
    delta_f_sq_yext = delta_f_sq_yp * ys + yp * delta_f_sq_ys

    dder = {"der_yext__f_sq": delta_f_sq_yext}

    return yext, dder 



import numpy
from typing import NoReturn

from cryspy.A_functions_base.function_1_objects import \
    form_items_by_dictionary

from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN


class PdPeak(ItemN):
    """Peak information calculated for diffraction pattern.
        
    This section contains peak information extracted from the
    measured or, if present, the processed diffractogram. Each
    peak in this table will have a unique label (see _pd_peak_id).
    The reflections and phases associated with each peak will be
    specified in other sections (see the _pd_refln_ and
    _pd_phase_ sections).

    Note that peak positions are customarily determined from the
    processed diffractogram and thus corrections for position
    and intensity will have been previously applied.

    Attributes:
        - index_h, index_k, index_l (mandatory)
        - index_mult, ttheta, intensity_plus, intensity_minus, width_ttheta
    """
    ATTR_MANDATORY_NAMES = ("index_h", "index_k", "index_l")
    ATTR_MANDATORY_TYPES = (int, int, int)
    ATTR_MANDATORY_CIF = ("index_h", "index_k", "index_l")

    ATTR_OPTIONAL_NAMES = (
        "index_mult", "ttheta", "intensity_plus", "intensity_minus",
        "width_ttheta", "sintlambda", "d_spacing", "gamma", "nu"
        )
    ATTR_OPTIONAL_TYPES = (
        int, float, float, float, float, float, float, float, float
        )
    ATTR_OPTIONAL_CIF = (
        "index_mult", "2theta", "intensity_plus",
        "intensity_minus", "width_2theta",
        "sint/lambda", "d_spacing", "gamma", "nu"
        )

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ()
    ATTR_INT_PROTECTED_NAMES = ()

    # parameters considered are refined parameters
    ATTR_REF = ()
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])
    ATTR_CONSTR_MARK = tuple([f"{_h:}_mark" for _h in ATTR_REF])

    # formats if cif format
    D_FORMATS = {
        "ttheta": "{:.2f}", "intensity_plus": "{:.2f}",
        "intensity_minus": "{:.2f}", "width_ttheta": "{:.5f}",
        "sintlambda": "{:.5f}", "d_spacing": "{:.5f}"}

    # constraints on the parameters
    D_CONSTRAINTS = {}

    # default values for the parameters
    D_DEFAULT = {}
    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False
    for key in ATTR_CONSTR_MARK:
        D_DEFAULT[key] = ""

    PREFIX = "pd_peak"

    def __init__(self, **kwargs) -> NoReturn:
        super(PdPeak, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {"sintlambda": 0., "d_spacing": 0.}

        # defined for ani integer and float parameters
        D_MAX = {}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)


class PdPeakL(LoopN):
    """Peaks calculated for diffraction pattern.

    This section contains peak information extracted from the
    measured or, if present, the processed diffractogram. Each
    peak in this table will have a unique label (see _pd_peak_id).
    The reflections and phases associated with each peak will be
    specified in other sections (see the _pd_refln_ and
    _pd_phase_ sections).

    Note that peak positions are customarily determined from the
    processed diffractogram and thus corrections for position
    and intensity will have been previously applied.

    Attributes:
        - index_h, index_k, index_l (mandatory)
        - index_mult, ttheta, intensity_plus, intensity_minus, width_ttheta

    """
    ITEM_CLASS = PdPeak
    ATTR_INDEX = None
    def __init__(self, loop_name: str = None, **kwargs) -> NoReturn:
        super(PdPeakL, self).__init__()
        self.__dict__["items"] = form_items_by_dictionary(self.ITEM_CLASS, kwargs)
        self.__dict__["loop_name"] = loop_name


# s_cont = """
#  loop_
#  _pd_peak_index_h
#  _pd_peak_index_k
#  _pd_peak_index_l
#  _pd_peak_index_mult
#  _pd_peak_2theta
#  _pd_peak_intensity_plus
#  _pd_peak_intensity_minus
#  _pd_peak_width_2theta
#   2  2  0  4 17.2 100.0  90.0  2.3
# """

# obj = PdPeakL.from_cif(s_cont)
# print(obj, end="\n\n")



import cryspy

from .mempy_by_dictionary import mempy_reconstruction_by_dictionary, mempy_cycle_density_susceptibility

def mempy_spin_density_reconstruction(obj: cryspy.GlobalN):
    if not(obj.is_attribute("mem_parameters")):
        mem_parameters = cryspy.MEMParameters(
            points_a=48, points_b=48, points_c=48,
            channel_plus_minus=True, magnetization_plus=4, magnetization_minus=-1)
        obj.add_items([mem_parameters,])
    obj.mem_parameters.channel_plus_minus = True
    obj.mem_parameters.channel_chi = False

    res = mempy_reconstruction_with_parameters(obj)
    return res 


def mempy_magnetization_density_reconstruction(obj: cryspy.GlobalN):
    if not(obj.is_attribute("mem_parameters")):
        mem_parameters = cryspy.MEMParameters(
            points_a=48, points_b=48, points_c=48, channel_chi=True, only_magnetic_basins=True)
        obj.add_items([mem_parameters,])
    obj.mem_parameters.channel_plus_minus = False
    obj.mem_parameters.channel_chi = True

    res = mempy_reconstruction_with_parameters(obj)
    return res  

def mempy_reconstruction_with_parameters(obj: cryspy.GlobalN,
        parameter_lambda:float=1.e-5, iteration_max:int=1000, parameter_lambda_min:float=1.e-9, delta_density:float=1.e-5):
    l_diffrn = [item for item in obj.items if isinstance(item, cryspy.Diffrn)]

    dict_obj = obj.get_dictionary()
    l_dict_crystal = [value for name, value in dict_obj.items() if name.startswith("crystal_")]
    l_dict_mem_parameters = [value for name, value in dict_obj.items() if name.startswith("mem_parameters")]
    l_dict_diffrn = [value for name, value in dict_obj.items() if name.startswith("diffrn_")]

    if len(l_dict_crystal) == 0:
        raise KeyError("The crystal is not found")

    elif len(l_dict_crystal) > 1:
        raise UserWarning("The first crystal will be used for MEM reconstruction procedure")

    if len(l_dict_diffrn) == 0:
        raise KeyError("The single crystall polarized experiments are not found")

    dict_crystal = l_dict_crystal[0]
    dict_mem_parameters = l_dict_mem_parameters[0]
    dict_in_out = {}
    mempy_reconstruction_by_dictionary(dict_crystal, dict_mem_parameters, l_dict_diffrn, dict_in_out,
        parameter_lambda=parameter_lambda, iteration_max=iteration_max,
        parameter_lambda_min=parameter_lambda_min, delta_density=delta_density)

    dict_in_out_keys = dict_in_out.keys()

    for diffrn in l_diffrn:
        flip_ratio = dict_in_out["dict_in_out_"+diffrn.get_name()]["flip_ratio"]
        diffrn_refln = diffrn.diffrn_refln
        diffrn_refln.numpy_fr_calc = flip_ratio
        diffrn_refln.numpy_to_items()

    if (("symm_elem_channel_chi" in dict_in_out_keys) and
            ("density_channel_chi" in dict_in_out_keys) and
            ("susceptibility_channel_chi" in dict_in_out_keys) and
            ("atom_multiplicity_channel_chi" in dict_in_out_keys) and
            ("point_multiplicity_channel_chi" in dict_in_out_keys)):

        symm_elem_channel_chi = dict_in_out["symm_elem_channel_chi"]  
        density_channel_chi = dict_in_out["density_channel_chi"]
        susceptibility_channel_chi = dict_in_out["susceptibility_channel_chi"]
        atom_multiplicity_channel_chi = dict_in_out["atom_multiplicity_channel_chi"]  
        point_multiplicity_channel_chi = dict_in_out["point_multiplicity_channel_chi"]  

        channel_chi = cryspy.ChannelChiL(loop_name=dict_crystal["name"])

        channel_chi.numpy_numerator_x = symm_elem_channel_chi[0]
        channel_chi.numpy_numerator_y = symm_elem_channel_chi[1]
        channel_chi.numpy_numerator_z = symm_elem_channel_chi[2]
        channel_chi.numpy_denominator_xyz = symm_elem_channel_chi[3]
        channel_chi.numpy_density = density_channel_chi
        channel_chi.numpy_chi_11 = susceptibility_channel_chi[0]
        channel_chi.numpy_chi_21 = susceptibility_channel_chi[1]
        channel_chi.numpy_chi_31 = susceptibility_channel_chi[2]
        channel_chi.numpy_chi_12 = susceptibility_channel_chi[3]
        channel_chi.numpy_chi_22 = susceptibility_channel_chi[4]
        channel_chi.numpy_chi_32 = susceptibility_channel_chi[5]
        channel_chi.numpy_chi_13 = susceptibility_channel_chi[6]
        channel_chi.numpy_chi_23 = susceptibility_channel_chi[7]
        channel_chi.numpy_chi_33 = susceptibility_channel_chi[8]
        channel_chi.numpy_atom_multiplicity = atom_multiplicity_channel_chi
        channel_chi.numpy_point_multiplicity = point_multiplicity_channel_chi
        channel_chi.numpy_to_items()
        obj.add_items([channel_chi, ])

    if (("symm_elem_channel_plus_minus" in dict_in_out_keys) and
            ("magnetization_plus" in dict_in_out_keys) and
            ("magnetization_minus" in dict_in_out_keys) and
            ("density_channel_plus_minus" in dict_in_out_keys) and
            ("point_multiplicity_channel_plus_minus" in dict_in_out_keys)):

        symm_elem_channel_plus_minus = dict_in_out["symm_elem_channel_plus_minus"]
        density_channel_plus_minus = dict_in_out["density_channel_plus_minus"]
        magnetization_plus = dict_in_out["magnetization_plus"]
        magnetization_minus = dict_in_out["magnetization_minus"]
        point_multiplicity_channel_plus_minus = dict_in_out["point_multiplicity_channel_plus_minus"]


        channel_plus_minus = cryspy.ChannelPlusMinusL(loop_name=dict_crystal["name"])

        channel_plus_minus.numpy_numerator_x = symm_elem_channel_plus_minus[0]
        channel_plus_minus.numpy_numerator_y = symm_elem_channel_plus_minus[1]
        channel_plus_minus.numpy_numerator_z = symm_elem_channel_plus_minus[2]
        channel_plus_minus.numpy_denominator_xyz = symm_elem_channel_plus_minus[3]
        channel_plus_minus.numpy_density_plus = density_channel_plus_minus[0]
        channel_plus_minus.numpy_density_minus = density_channel_plus_minus[1]
        channel_plus_minus.numpy_point_multiplicity = point_multiplicity_channel_plus_minus
        channel_plus_minus.numpy_to_items()
        obj.add_items([channel_plus_minus, ])
    return 



def mempy_cycle_with_parameters(obj: cryspy.GlobalN,
        parameter_lambda:float=1.e-5, iteration_max:int=1000, parameter_lambda_min:float=1.e-9, delta_density:float=1.e-5,
        n_cycle:int=10):

    l_diffrn = [item for item in obj.items if isinstance(item, cryspy.Diffrn)]

    dict_obj = obj.get_dictionary()
    l_dict_crystal = [value for name, value in dict_obj.items() if name.startswith("crystal_")]
    l_dict_mem_parameters = [value for name, value in dict_obj.items() if name.startswith("mem_parameters")]
    l_dict_diffrn = [value for name, value in dict_obj.items() if name.startswith("diffrn_")]

    if len(l_dict_crystal) == 0:
        raise KeyError("The crystal is not found")

    elif len(l_dict_crystal) > 1:
        raise UserWarning("The first crystal will be used for MEM reconstruction procedure")

    if len(l_dict_diffrn) == 0:
        raise KeyError("The single crystall polarized experiments are not found")

    dict_crystal = l_dict_crystal[0]
    dict_mem_parameters = l_dict_mem_parameters[0]
    

    dict_in_out = {}
    mempy_cycle_density_susceptibility(dict_crystal, dict_mem_parameters, l_dict_diffrn, dict_in_out,
        parameter_lambda=parameter_lambda, iteration_max=iteration_max, parameter_lambda_min=parameter_lambda_min,
        delta_density=delta_density, n_cycle=n_cycle)

    dict_in_out_den = dict_in_out["dict_in_out_den"]
    dict_in_out_susc = dict_in_out["dict_in_out_susc"]

    if "atom_para_susceptibility_sigma" in dict_in_out_susc.keys():
        atom_para_susceptibility_sigma = dict_in_out_susc["atom_para_susceptibility_sigma"]
        atom_para_susceptibility = dict_crystal["atom_para_susceptibility"]
        flags_atom_para_susceptibility = dict_crystal["flags_atom_para_susceptibility"]
        # obj.take_parameters_from_dictionary(dict_obj, l_parameter_name = parameter_name, l_sigma=sigma_p)
        #FIXME: temporary solution
        obj_crystal = getattr(obj, f"crystal_{dict_crystal['name']:}")
        atom_site_susceptibility = obj_crystal.atom_site_susceptibility
        l_attr = ["chi_11", "chi_22", "chi_33", "chi_12", "chi_13", "chi_23"]
        for i_item, item in enumerate(atom_site_susceptibility.items):
            for i_attr, attr in enumerate(l_attr):
                setattr(item, attr, atom_para_susceptibility[i_attr, i_item])
                if flags_atom_para_susceptibility[i_attr, i_item]:
                    setattr(item, f"{attr:}_sigma", atom_para_susceptibility_sigma[i_attr, i_item])


    dict_in_out_den_keys = dict_in_out_den.keys()

    for diffrn in l_diffrn:
        flip_ratio = dict_in_out_den["dict_in_out_"+diffrn.get_name()]["flip_ratio"]
        diffrn_refln = diffrn.diffrn_refln
        diffrn_refln.numpy_fr_calc = flip_ratio
        diffrn_refln.numpy_to_items()

    if (("symm_elem_channel_chi" in dict_in_out_den_keys) and
            ("density_channel_chi" in dict_in_out_den_keys) and
            ("susceptibility_channel_chi" in dict_in_out_den_keys) and
            ("atom_multiplicity_channel_chi" in dict_in_out_den_keys) and
            ("point_multiplicity_channel_chi" in dict_in_out_den_keys)):

        symm_elem_channel_chi = dict_in_out_den["symm_elem_channel_chi"]  
        density_channel_chi = dict_in_out_den["density_channel_chi"]
        susceptibility_channel_chi = dict_in_out_den["susceptibility_channel_chi"]
        atom_multiplicity_channel_chi = dict_in_out_den["atom_multiplicity_channel_chi"]  
        point_multiplicity_channel_chi = dict_in_out_den["point_multiplicity_channel_chi"]  

        channel_chi = cryspy.ChannelChiL(loop_name=dict_crystal["name"])

        channel_chi.numpy_numerator_x = symm_elem_channel_chi[0]
        channel_chi.numpy_numerator_y = symm_elem_channel_chi[1]
        channel_chi.numpy_numerator_z = symm_elem_channel_chi[2]
        channel_chi.numpy_denominator_xyz = symm_elem_channel_chi[3]
        channel_chi.numpy_density = density_channel_chi
        channel_chi.numpy_chi_11 = susceptibility_channel_chi[0]
        channel_chi.numpy_chi_21 = susceptibility_channel_chi[1]
        channel_chi.numpy_chi_31 = susceptibility_channel_chi[2]
        channel_chi.numpy_chi_12 = susceptibility_channel_chi[3]
        channel_chi.numpy_chi_22 = susceptibility_channel_chi[4]
        channel_chi.numpy_chi_32 = susceptibility_channel_chi[5]
        channel_chi.numpy_chi_13 = susceptibility_channel_chi[6]
        channel_chi.numpy_chi_23 = susceptibility_channel_chi[7]
        channel_chi.numpy_chi_33 = susceptibility_channel_chi[8]
        channel_chi.numpy_atom_multiplicity = atom_multiplicity_channel_chi
        channel_chi.numpy_point_multiplicity = point_multiplicity_channel_chi
        channel_chi.numpy_to_items()
        obj.add_items([channel_chi, ])

    if (("symm_elem_channel_plus_minus" in dict_in_out_den_keys) and
            ("magnetization_plus" in dict_in_out_den_keys) and
            ("magnetization_minus" in dict_in_out_den_keys) and
            ("density_channel_plus_minus" in dict_in_out_den_keys) and
            ("point_multiplicity_channel_plus_minus" in dict_in_out_den_keys)):

        symm_elem_channel_plus_minus = dict_in_out_den["symm_elem_channel_plus_minus"]
        density_channel_plus_minus = dict_in_out_den["density_channel_plus_minus"]
        magnetization_plus = dict_in_out_den["magnetization_plus"]
        magnetization_minus = dict_in_out_den["magnetization_minus"]
        point_multiplicity_channel_plus_minus = dict_in_out_den["point_multiplicity_channel_plus_minus"]


        channel_plus_minus = cryspy.ChannelPlusMinusL(loop_name=dict_crystal["name"])

        channel_plus_minus.numpy_numerator_x = symm_elem_channel_plus_minus[0]
        channel_plus_minus.numpy_numerator_y = symm_elem_channel_plus_minus[1]
        channel_plus_minus.numpy_numerator_z = symm_elem_channel_plus_minus[2]
        channel_plus_minus.numpy_denominator_xyz = symm_elem_channel_plus_minus[3]
        channel_plus_minus.numpy_density_plus = density_channel_plus_minus[0]
        channel_plus_minus.numpy_density_minus = density_channel_plus_minus[1]
        channel_plus_minus.numpy_point_multiplicity = point_multiplicity_channel_plus_minus
        channel_plus_minus.numpy_to_items()
        obj.add_items([channel_plus_minus, ])
    print(70*"*")
    print("End of cycle procedure (MEMPy module)")
    print(70*"*")



"""Transform data to objects of cryspy library."""
import os
import os.path
from typing import List

from pycifstar import Global, to_global, Items, Loop, Data

from cryspy.A_functions_base.function_1_strings import find_prefix

from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN
from cryspy.B_parent_classes.cl_3_data import DataN
from cryspy.B_parent_classes.cl_4_global import GlobalN

from cryspy.C_item_loop_classes.cl_1_atom_site import AtomSite, \
    AtomSiteL
from cryspy.C_item_loop_classes.cl_1_atom_type import AtomType, \
    AtomTypeL
from cryspy.C_item_loop_classes.cl_1_atom_site_aniso import \
    AtomSiteAniso, AtomSiteAnisoL
from cryspy.C_item_loop_classes.cl_1_atom_site_moment import \
    AtomSiteMoment, AtomSiteMomentL
from cryspy.C_item_loop_classes.cl_1_atom_site_susceptibility import \
    AtomSiteSusceptibility, AtomSiteSusceptibilityL
from cryspy.C_item_loop_classes.cl_1_atom_site_exchange import \
    AtomSiteExchange, AtomSiteExchangeL
from cryspy.C_item_loop_classes.cl_1_atom_type_scat import \
    AtomTypeScat, AtomTypeScatL
from cryspy.C_item_loop_classes.cl_1_atom_local_axes import \
    AtomLocalAxes, AtomLocalAxesL
from cryspy.C_item_loop_classes.cl_1_atom_electron_configuration \
    import AtomElectronConfiguration, AtomElectronConfigurationL
from cryspy.C_item_loop_classes.cl_1_cell import Cell, CellL
from cryspy.C_item_loop_classes.cl_1_channel_chi import ChannelChi, ChannelChiL
from cryspy.C_item_loop_classes.cl_1_channel_plus_minus import ChannelPlusMinus, ChannelPlusMinusL
from cryspy.C_item_loop_classes.cl_1_chi2 import Chi2, Chi2L
from cryspy.C_item_loop_classes.cl_1_diffrn_radiation import \
    DiffrnRadiation, DiffrnRadiationL
from cryspy.C_item_loop_classes.cl_1_diffrn_refln import DiffrnRefln, \
    DiffrnReflnL
from cryspy.C_item_loop_classes.cl_1_exclude import Exclude, ExcludeL
from cryspy.C_item_loop_classes.cl_1_extinction import Extinction, \
    ExtinctionL
from cryspy.C_item_loop_classes.cl_1_inversed_hessian import InversedHessian
from cryspy.C_item_loop_classes.cl_1_mem_parameters import \
    MEMParameters, MEMParametersL
from cryspy.C_item_loop_classes.cl_1_phase import Phase, PhaseL
from cryspy.C_item_loop_classes.cl_1_pd2d_instr_reflex_asymmetry import\
    Pd2dInstrReflexAsymmetry, Pd2dInstrReflexAsymmetryL
from cryspy.C_item_loop_classes.cl_1_pd2d_instr_resolution import \
    Pd2dInstrResolution, Pd2dInstrResolutionL
from cryspy.C_item_loop_classes.cl_1_pd2d_meas import Pd2dMeas
from cryspy.C_item_loop_classes.cl_1_pd2d_proc import Pd2dProc
from cryspy.C_item_loop_classes.cl_1_pd2d_peak import Pd2dPeak, \
    Pd2dPeakL
from cryspy.C_item_loop_classes.cl_1_pd2d_background import \
    Pd2dBackground
from cryspy.C_item_loop_classes.cl_1_pd_background import \
    PdBackground, PdBackgroundL
from cryspy.C_item_loop_classes.cl_1_pd_instr_reflex_asymmetry import \
    PdInstrReflexAsymmetry, PdInstrReflexAsymmetryL
from cryspy.C_item_loop_classes.cl_1_pd_instr_resolution import \
    PdInstrResolution, PdInstrResolutionL
from cryspy.C_item_loop_classes.cl_1_pd_meas import PdMeas, PdMeasL
from cryspy.C_item_loop_classes.cl_1_pd_proc import PdProc, PdProcL
from cryspy.C_item_loop_classes.cl_1_pd_peak import PdPeak, PdPeakL
from cryspy.C_item_loop_classes.cl_1_range import Range, RangeL
from cryspy.C_item_loop_classes.cl_1_refln import Refln, ReflnL
from cryspy.C_item_loop_classes.cl_1_refln_susceptibility import \
    ReflnSusceptibility, ReflnSusceptibilityL
from cryspy.C_item_loop_classes.cl_1_refine_ls import RefineLs, \
    RefineLsL
from cryspy.C_item_loop_classes.cl_1_setup import Setup, SetupL
from cryspy.C_item_loop_classes.cl_1_space_group_symop import \
    SpaceGroupSymop, SpaceGroupSymopL
from cryspy.C_item_loop_classes.cl_1_space_group_symop_magn_centering import \
    SpaceGroupSymopMagnCentering, SpaceGroupSymopMagnCenteringL
from cryspy.C_item_loop_classes.cl_1_space_group_wyckoff import \
    SpaceGroupWyckoff, SpaceGroupWyckoffL
from cryspy.C_item_loop_classes.cl_1_texture import Texture, TextureL
from cryspy.C_item_loop_classes.cl_1_tof_background import TOFBackground, \
    TOFBackgroundL
from cryspy.C_item_loop_classes.cl_1_tof_intensity_incident import \
    TOFIntensityIncident, TOFIntensityIncidentL
from cryspy.C_item_loop_classes.cl_1_tof_meas import TOFMeas, TOFMeasL
from cryspy.C_item_loop_classes.cl_1_tof_parameters import TOFParameters, \
    TOFParametersL
from cryspy.C_item_loop_classes.cl_1_tof_peak import TOFPeak, TOFPeakL
from cryspy.C_item_loop_classes.cl_1_tof_proc import TOFProc, TOFProcL
from cryspy.C_item_loop_classes.cl_1_tof_profile import TOFProfile, TOFProfileL


from cryspy.C_item_loop_classes.cl_2_atom_rho_orbital_radial_slater import \
    AtomRhoOrbitalRadialSlater, AtomRhoOrbitalRadialSlaterL
from cryspy.C_item_loop_classes.cl_2_atom_site_scat import AtomSiteScat, \
    AtomSiteScatL
from cryspy.C_item_loop_classes.cl_2_diffrn_orient_matrix import \
    DiffrnOrientMatrix, DiffrnOrientMatrixL
from cryspy.C_item_loop_classes.cl_2_section import Section, SectionL
from cryspy.C_item_loop_classes.cl_2_space_group import SpaceGroup
from cryspy.C_item_loop_classes.cl_2_space_group_symop_magn_operation import \
    SpaceGroupSymopMagnOperation, SpaceGroupSymopMagnOperationL

from cryspy.C_item_loop_classes.cl_3_density_point import DensityPoint, \
    DensityPointL

from cryspy.E_data_classes.cl_1_crystal import Crystal
# from cryspy.E_data_classes.cl_1_mag_crystal import MagCrystal
from cryspy.E_data_classes.cl_2_diffrn import Diffrn
from cryspy.E_data_classes.cl_2_pd import Pd
from cryspy.E_data_classes.cl_2_pd2d import Pd2d
from cryspy.E_data_classes.cl_2_tof import TOF

L_ITEM_CLASS = []
L_LOOP_CLASS = []
L_DATA_CLASS = []
L_GLOBAL_CLASS = []
L_FUNCTION_ADD = []


L_ITEM_CLASS.extend([
    AtomElectronConfiguration, AtomLocalAxes, AtomRhoOrbitalRadialSlater,
    AtomSite, AtomSiteAniso, AtomSiteMoment, AtomSiteScat,
    AtomSiteSusceptibility, AtomSiteExchange, AtomType, AtomTypeScat,
    Cell, ChannelChi, ChannelPlusMinus, Chi2,
    DensityPoint, DiffrnRadiation, DiffrnOrientMatrix, DiffrnRefln,
    Exclude, Extinction, MEMParameters, InversedHessian,
    Phase, PdBackground,
    PdInstrReflexAsymmetry, PdInstrResolution, PdMeas, PdProc, PdPeak,
    Pd2dBackground, Pd2dInstrReflexAsymmetry,
    Pd2dInstrResolution, Pd2dMeas, Pd2dProc, Pd2dPeak,
    Range, RefineLs, Refln, ReflnSusceptibility,
    Section, Setup, SpaceGroupSymopMagnCentering, SpaceGroupSymopMagnOperation,
    SpaceGroup, SpaceGroupSymop, SpaceGroupWyckoff, Texture, TOFBackground,
    TOFIntensityIncident, TOFMeas,
    TOFParameters, TOFPeak, TOFProc, TOFProfile])

L_LOOP_CLASS.extend([
    AtomLocalAxesL, AtomElectronConfigurationL, AtomRhoOrbitalRadialSlaterL,
    AtomSiteL, AtomSiteAnisoL, AtomSiteMomentL, AtomSiteScatL,
    AtomSiteSusceptibilityL, AtomSiteExchangeL, AtomTypeL, AtomTypeScatL,
    CellL, ChannelChiL, ChannelPlusMinusL, Chi2L,
    DensityPointL, DiffrnRadiationL, DiffrnOrientMatrixL, DiffrnReflnL,
    ExtinctionL, ExcludeL, MEMParametersL,
    PhaseL, PdBackgroundL, PdInstrReflexAsymmetryL, PdInstrResolutionL,
    PdMeasL, PdProcL, PdPeakL,
    Pd2dInstrReflexAsymmetryL, Pd2dInstrResolutionL, Pd2dPeakL,
    RangeL, RefineLsL, ReflnL, ReflnSusceptibilityL,
    SectionL, SetupL, SpaceGroupSymopL, SpaceGroupSymopMagnCenteringL,
    SpaceGroupSymopMagnOperationL, SpaceGroupWyckoffL, TextureL,
    TOFBackgroundL, TOFIntensityIncidentL, TOFMeasL, TOFParametersL, TOFPeakL,
    TOFProcL, TOFProfileL])

L_DATA_CLASS.extend([Crystal, Diffrn, Pd, Pd2d, TOF]) # , MagCrystal

L_GLOBAL_CLASS.extend([])

F_PACKAGES = os.path.join(os.path.dirname(__file__), "packages.dat")

import sys
from importlib import import_module


def load_packages():
    """Load external packages."""
    if os.path.isfile(F_PACKAGES):
        with open(F_PACKAGES, "r") as fid:
            l_content = fid.readlines()
    else:
        return
    l_content_new = []
    for f_name in l_content:
        f_name_sh = f_name.strip()
        f_ok = True
        if os.path.isdir(f_name.strip()):
            module_name = os.path.basename(f_name_sh)
            module_way = os.path.dirname(f_name_sh)
            sys.path.append(module_way)
            module = import_module(module_name)
            try:
                for obj in module.CRYSPY_CLASSES:
                    if issubclass(obj, ItemN):
                        if not(obj in L_ITEM_CLASS):
                            L_ITEM_CLASS.append(obj)
                    elif issubclass(obj, LoopN):
                        if not(obj in L_LOOP_CLASS):
                            L_LOOP_CLASS.append(obj)
                    elif issubclass(obj, DataN):
                        if not(obj in L_DATA_CLASS):
                            L_DATA_CLASS.append(obj)
                    elif issubclass(obj, GlobalN):
                        if not(obj in L_GLOBAL_CLASS):
                            L_GLOBAL_CLASS.append(obj)
            except AttributeError:
                print(f"Module '{module_name:}' in folder: \n {module_way}\n\
does not contains in 'init.py' parameter 'CRYSPY_CLASSES'.")
                f_ok = False
            try:
                L_FUNCTION_ADD.extend(module.L_FUNCTION)
            except AttributeError:
                print(f"Module '{module_name:}' in folder: \n {module_way}\n\
does not contains in 'init.py' parameter 'L_FUNCTION'.")
                f_ok = False
        else:
            f_ok = False
        if f_ok:
            l_content_new.append(f_name_sh)
    with open(F_PACKAGES, "w") as fid:
        fid.write("\n".join(l_content_new)+"\n")


def add_package(f_name: str):
    """Add packages to cryspy library."""
    if os.path.isfile(F_PACKAGES):
        with open(F_PACKAGES, 'r') as fid:
            l_cont = fid.readlines()
        if not(f_name+"\n" in l_cont):
            with open(F_PACKAGES, 'a') as fid:
                fid.write(f_name+"\n")
    else:
        with open(F_PACKAGES, 'a') as fid:
            fid.write(f_name+"\n")


def packages():
    """Give list of installed packages."""
    if os.path.isfile(F_PACKAGES):
        with open(F_PACKAGES, "r") as fid:
            l_content = fid.readlines()
        l_res = [hh for hh in l_content if hh.strip() != ""]
    else:
        l_res = []
    return l_res


def delete_package(f_name: str):
    """Delete from the list of installed packages."""
    f_name_sh = f_name.strip()
    if os.path.isfile(F_PACKAGES):
        with open(F_PACKAGES, "r") as fid:
            l_content = fid.readlines()
        l_res = [hh for hh in l_content if hh.strip() != f_name_sh]
        with open(F_PACKAGES, "w") as fid:
            fid.write("\n".join(l_res) + "\n")


def is_in_rcif_block(rcif_block, cryspy_obj):
    """Is in rcif block."""
    ls_cryspy = cryspy_obj.to_cif.split("\n")
    l_name = [_.strip().split()[0] for _ in ls_cryspy if _.startswith("_")]
    l_flag = [rcif_block.is_value(_name) for _name in l_name]
    return any(l_flag), all(l_flag)


def file_to_globaln(file_name: str, item_classes=(), loop_classes=(),
                    data_classes=(), global_classes=()) -> GlobalN:
    """Transfer file content of cif format to cryspy objects."""
    if not(os.path.isfile(file_name)):
        raise UserWarning(f"File '{file_name:}' is not found")
        return
    s_cont = str(to_global(file_name))
    res = str_to_globaln(s_cont, item_classes=item_classes,
                         loop_classes=loop_classes, data_classes=data_classes,
                         global_classes=global_classes)
    return res


def str_to_items(s_cont: str, item_classes=(), loop_classes=()) -> list:
    """Transfer string of cif format to cryspy objects."""
    data_cif = Data()
    data_cif.take_from_string(s_cont)
    l_item_class = L_ITEM_CLASS + list(item_classes)
    l_loop_class = L_LOOP_CLASS + list(loop_classes)

    l_item = items_to_itemsn(data_cif.items, l_item_class)
    for loop_cif in data_cif.loops:
        l_loop_obj = loop_to_loopn(loop_cif, l_loop_class)
        l_item.extend(l_loop_obj)
    return l_item


def str_to_globaln(s_cont: str, item_classes=(), loop_classes=(),
                   data_classes=(), global_classes=()) -> GlobalN:
    """Transfer string of cif format to cryspy objects."""
    global_cif = Global()
    global_cif.take_from_string(s_cont)
    l_global_item = []
    l_cls_global = []
    l_item_class = L_ITEM_CLASS + list(item_classes)
    l_loop_class = L_LOOP_CLASS + list(loop_classes)
    l_data_class = L_DATA_CLASS + list(data_classes)
    l_global_class = L_GLOBAL_CLASS + list(global_classes)
    # l_global_class = []
    if len(global_cif.items)+len(global_cif.loops) > 0:
        l_item_obj = items_to_itemsn(global_cif.items, l_item_class)
        l_global_item.extend(l_item_obj)

        for item in l_item_obj:
            if type(item) not in l_cls_global:
                l_cls_global.append(type(item))

        for loop_cif in global_cif.loops:
            l_loop_obj = loop_to_loopn(loop_cif, l_loop_class)
            l_global_item.extend(l_loop_obj)
            for loop_obj in l_loop_obj:
                if type(loop_obj) not in l_cls_global:
                    l_cls_global.append(type(loop_obj))

    for data_cif in global_cif.datas:
        flag = False
        str_data = str(data_cif)

        for cls_data in l_data_class:
            # FIXME: it's bad solution as we loose information which are
            #        not specified as knonw in cryspy library.
            mand_line = cls_data.get_mandatory_attributes(separator="_")
            mand_point = cls_data.get_mandatory_attributes(separator=".")
            l_flag = []
            for m_l, m_p in zip(mand_line, mand_point):
                flag_t = data_cif.is_value(m_l)
                if not(flag_t):
                    flag_t = data_cif.is_value(m_p)
                l_flag.append(flag_t)
            if all(l_flag):
                #print("all(l_flag): \n", all(l_flag), end="\n\n")
                data_obj = cls_data.from_cif(str_data)
                if data_obj is not None:
                    l_global_item.append(data_obj)
                    if not(cls_data in l_cls_global):
                        l_cls_global.append(cls_data)
                    flag = True
        if not(flag):
            l_data_item = []
            l_cls_data = []

            l_item_obj = items_to_itemsn(data_cif.items, l_item_class)
            l_data_item.extend(l_item_obj)
            for item in l_item_obj:
                if type(item) not in l_cls_data:
                    l_cls_data.append(type(item))

            for loop_cif in data_cif.loops:
                l_loop_obj = loop_to_loopn(loop_cif, l_loop_class)
                l_data_item.extend(l_loop_obj)
                for loop_obj in l_loop_obj:
                    if type(loop_obj) not in l_cls_data:
                        l_cls_data.append(type(loop_obj))

            if len(l_data_item) > 0:
                data_obj = items_to_datan(data_cif.name, l_data_item,
                                          l_data_class)
                l_global_item.append(data_obj)


    global_obj = items_to_globaln(global_cif.name, l_global_item,
                                  l_global_class)
    return global_obj


def find_name_prefix(loop_name, l_item_cls):
    """Find name prefix."""
    loop_name = loop_name.strip("_")
    l_loop_prefix = [item_cls.PREFIX for item_cls in l_item_cls]
    l_loop = [loop_prefix for loop_prefix in l_loop_prefix if
              loop_name.startswith(loop_prefix)]
    if len(l_loop) > 0:
        l_loop.sort(key=lambda x: len(x))
        for prefix in l_loop:
            name = loop_name.strip("_")[len(prefix)+1:]
            item_cls = l_item_cls[l_loop_prefix.index(prefix)]
            l_attr_cif = [attr.lower() for attr in item_cls.ATTR_CIF]
            if name.lower() in l_attr_cif:
                separator = loop_name.strip("_")[len(prefix)]
                return prefix, name, separator
        # prefix = l_loop[-1]
        # name = loop_name.strip("_")[len(prefix)+1:]
        # separator = loop_name.strip("_")[len(prefix)]
        # return prefix, name, separator
    return None, None, None


# TODO: l_item_class should be added as not necessary l_loop_class fully
#       correspond to l_item_class
def loop_to_loopn(loop_cif: Loop, l_loop_class: list) -> LoopN:
    """Transform Loop object to ItemN objects."""
    # l_item_class for future.
    l_item_class = []
    l_item_prefix = [item_cls.PREFIX for item_cls in l_item_class]
    l_item_cls_in_def_loop = [loop_cls.ITEM_CLASS for loop_cls in l_loop_class]
    l_loop_prefix = [loop_cls.ITEM_CLASS.PREFIX for loop_cls in l_loop_class]
    l_loop_name = [name.lower() for name in loop_cif.names]
    flag_id = False
    flag_label = False
    flag_type = False
    l_loop_prefix_in = []
    l_loop_name_in = []
    l_loop_separator_in = []

    for loop_name in l_loop_name:
        prefix, name, separator = find_name_prefix(loop_name,
                                                   l_item_cls_in_def_loop)

        if prefix is not None:
            l_name_allowed = [h.lower() for h in l_loop_class[
                l_loop_prefix.index(prefix)].ITEM_CLASS.ATTR_CIF]
            if not(name.lower() in l_name_allowed):
                prefix = None
                name = None

        if (prefix is None) | (name is None):
            prefix, name, separator = find_name_prefix(loop_name,
                                                       l_item_cls_in_def_loop)
            if prefix is not None:
                l_name_allowed = [h.lower() for h in l_item_class[
                    l_item_prefix.index(prefix)].ATTR_CIF]
                if not(name.lower() in l_name_allowed):
                    prefix = None
                    name = None

        if (prefix is None) | (name is None):
            if loop_name.find(".") != -1:
                prefix, name = loop_name.strip("_").split(".")[:2]
                separator = "."
            else:
                prefix = loop_name.strip("_").split("_")[0]
                name = loop_name.strip("_")[len(prefix)+1:]
                separator = "_"

        if name.lower() == "id":
            flag_id = True
            prefix_id = prefix
            loop_name_id = loop_name

        if name.lower() == "label":
            flag_label = True
            prefix_label = prefix
            loop_name_label = loop_name

        if name.lower() == "type":
            flag_type = True
            prefix_type = prefix
            loop_name_type = loop_name

        l_loop_prefix_in.append(prefix)
        l_loop_name_in.append(name)
        l_loop_separator_in.append(separator)


    s_loop_prefix_in = set(l_loop_prefix_in)
    l_loopn = []
    for prefix in s_loop_prefix_in:
        item_cls = ItemN
        loop_cls = LoopN
        if prefix in l_loop_prefix:
            loop_cls = l_loop_class[l_loop_prefix.index(prefix)]
        elif prefix in l_item_prefix:
            # for future
            item_cls = l_item_class[l_loop_prefix.index(prefix)]

        l_loop_name_cif = []
        ll_val_cif = []
        separator = "_"
        for loop_name, prefix_in, name_in, separator_in in zip(
                l_loop_name, l_loop_prefix_in, l_loop_name_in,
                l_loop_separator_in):
            if prefix == prefix_in:
                l_loop_name_cif.append(loop_name)
                ll_val_cif.append(loop_cif[loop_name])
                separator = separator_in
        if flag_id:
            if prefix_id != prefix:
                loop_new_name_id = "_"+prefix+f"{separator:}id"
                l_loop_name_cif.append(loop_new_name_id)
                ll_val_cif.append(loop_cif[loop_name_id])
        elif flag_label:
            if prefix_label != prefix:
                loop_new_name_label = "_"+prefix+f"{separator:}label"
                l_loop_name_cif.append(loop_new_name_label)
                ll_val_cif.append(loop_cif[loop_name_label])
        elif flag_type:
            if prefix_type != prefix:
                loop_new_name_type = "_"+prefix+f"{separator:}type"
                l_loop_name_cif.append(loop_new_name_type)
                ll_val_cif.append(loop_cif[loop_name_type])
        ll_val_cif_t = [[ll_val_cif[i_2][i_1] for i_2 in range(len(
            ll_val_cif))] for i_1 in range(len(ll_val_cif[0]))]
        obj = Loop(names=l_loop_name_cif, values=ll_val_cif_t, name=loop_cif.name)
        loopn = loop_cls.from_cif(str(obj))

        if loopn is None:
            loop_cls = LoopN
            loopn = loop_cls.from_cif(str(obj))
        # if len(loopn.items) == 0:
        #     print("obj:", loop_cls)
        #     print(obj)
        if not(loopn is None):
            l_loopn.append(loopn)
        else:
            print("l_loop_name_cif: \n", l_loop_name_cif)
    return l_loopn


def items_to_itemsn(items_cif: Items, l_item_class: list) -> List[ItemN]:
    """Transform Items object to items of ItemN classes."""

    l_item = []
    l_name_cif = list(items_cif.names)
    flag_point_separator = all([name.find(".") == -1
                                for name in l_name_cif])
    if flag_point_separator:
        separator = "_"
    else:
        separator = "."

    for cls_item in l_item_class:
        prefix = cls_item.PREFIX

        separator_loc = "_"

        l_cls_mand_cif = [f"_{prefix:}{separator_loc:}{attr_cif:}"
                          for attr_cif in cls_item.ATTR_MANDATORY_CIF]

        l_cls_all_cif = [f"_{prefix:}{separator_loc:}{attr_cif:}"
                         for attr_cif in (cls_item.ATTR_MANDATORY_CIF +
                                          cls_item.ATTR_OPTIONAL_CIF)]

        flag_mand = all([items_cif.is_value(name) for name in l_cls_mand_cif])
        flag_any = any([items_cif.is_value(name) for name in l_cls_all_cif])

        if not(flag_any):
            separator_loc = "."
            l_cls_mand_cif = [f"_{prefix:}{separator_loc:}{attr_cif:}"
                              for attr_cif in cls_item.ATTR_MANDATORY_CIF]

            l_cls_all_cif = [f"_{prefix:}{separator_loc:}{attr_cif:}"
                             for attr_cif in (cls_item.ATTR_MANDATORY_CIF +
                                              cls_item.ATTR_OPTIONAL_CIF)]

            flag_mand = all([items_cif.is_value(name) for name in
                             l_cls_mand_cif])
            flag_any = any([items_cif.is_value(name) for name in
                            l_cls_all_cif])

        item_obj = None
        if flag_mand:
            items_small = items_cif.items_with_prefix(f"_{prefix:}")
            item_obj = cls_item.from_cif(str(items_small))
        elif flag_any:
            items_small = items_cif.items_with_prefix(f"_{prefix:}")
            item_obj = ItemN.from_cif(str(items_small))

        if item_obj is not None:
            l_item.append(item_obj)
            l_name_cif_remove = [name.lower() for name in l_cls_all_cif
                                 if bool(l_name_cif.count(name.lower()))]
            for name in l_name_cif_remove:
                l_name_cif.remove(name)

    if len(l_name_cif) != 0:
        if separator == ".":
            l_prefix_cif = [(name[1:]).split(separator)[0]
                            for name in l_name_cif]
        else:
            l_prefix_cif = find_prefixes(l_name_cif)

        for prefix in l_prefix_cif:
            names = [name_cif for name_cif in l_name_cif if
                     name_cif.startswith("_"+prefix)]
            items_small = items_cif.items_with_names(names)
            item_obj = ItemN.from_cif(str(items_small))
            l_item.append(item_obj)
    return l_item


def find_prefixes(l_name_cif):
    """Find prefix."""
    l_short_prefix_cif = list(set([(name[1:]).split("_")[0]
                                   for name in l_name_cif]))
    l_prefix_cif = []
    for short_prefix_cif in l_short_prefix_cif:
        l_name_2_cif = [name for name in l_name_cif
                        if (name[1:]).split("_")[0] == short_prefix_cif]
        prefix_cif = find_prefix(*tuple(l_name_2_cif))
        l_prefix_cif.append(prefix_cif[1:])
    return l_prefix_cif


def items_to_globaln(global_name: str, items: list, l_global_class):
    """
    Items to suitable GlobalN class.

    Parameters
    ----------
    global_name : str
        name of the GlobalN.
    items : List[Union[ItemN, LoopN, DataN]]
        List of items.
    l_global_class : list
        List of GlobalN classes to choose one of them.

    Returns
    -------
    global_obj : TYPE
        Object of GlobalN class or one of l_global_classes.

    """
    global_obj = None
    classes = set([type(item) for item in items])
    l_global_class  = []
    for cls_global in l_global_class:
        flag_mand = all([cls_mand in classes
                         for cls_mand in cls_global.CLASSES_MANDATORY])
        flag_opt = all([class_ in cls_global.CLASSES
                        for class_ in classes])
        if (flag_mand & flag_opt):
            global_obj = cls_global()
            global_obj.global_name = global_name
            global_obj.add_items(items)

    if global_obj is None:
        global_obj = GlobalN.make_container((), tuple(classes), "global")
        global_obj.global_name = global_name
        global_obj.add_items(items)
    return global_obj


def items_to_datan(data_name: str, items: list, l_data_class):
    """
    Items to suitable GlobalN class.

    Parameters
    ----------
    data_name : str
        name of the DataN.
    items : List[Union[ItemN, LoopN]]
        List of items.
    l_data_class : list
        List of DataN classes to choose one of them.

    Returns
    -------
    data_obj : TYPE
        Object of DataN class or one of l_data_classes.

    """
    data_obj = None
    classes = set([type(item) for item in items])
    for cls_data in l_data_class:
        flag_mand = all([cls_mand in classes
                         for cls_mand in cls_data.CLASSES_MANDATORY])
        flag_opt = all([class_ in cls_data.CLASSES
                        for class_ in classes])

        if (flag_mand & flag_opt):
            data_obj = cls_data()
            data_obj.data_name = data_name
            data_obj.add_items(items)

    if data_obj is None:
        data_obj = DataN.make_container((), tuple(classes), "data")
        data_obj.data_name = data_name
        data_obj.add_items(items)
    return data_obj



from typing import NoReturn
import numpy

from cryspy.A_functions_base.function_1_objects import \
    form_items_by_dictionary

from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN


class PdInstrReflexAsymmetry(ItemN):
    """Asymmetry of Bragg reflections for 1d powder diffractometer.
    """
    ATTR_MANDATORY_NAMES = ("p1", "p2", "p3", "p4")
    ATTR_MANDATORY_TYPES = (float, float, float, float)
    ATTR_MANDATORY_CIF = ("p1", "p2", "p3", "p4")

    ATTR_OPTIONAL_NAMES = ()
    ATTR_OPTIONAL_TYPES = ()
    ATTR_OPTIONAL_CIF = ()

    ATTR_NAMES = ATTR_MANDATORY_NAMES + ATTR_OPTIONAL_NAMES
    ATTR_TYPES = ATTR_MANDATORY_TYPES + ATTR_OPTIONAL_TYPES
    ATTR_CIF = ATTR_MANDATORY_CIF + ATTR_OPTIONAL_CIF

    ATTR_INT_NAMES = ()
    ATTR_INT_PROTECTED_NAMES = ()

    # parameters considered are refined parameters
    ATTR_REF = ("p1", "p2", "p3", "p4")
    ATTR_SIGMA = tuple([f"{_h:}_sigma" for _h in ATTR_REF])
    ATTR_CONSTR_FLAG = tuple([f"{_h:}_constraint" for _h in ATTR_REF])
    ATTR_REF_FLAG = tuple([f"{_h:}_refinement" for _h in ATTR_REF])
    ATTR_CONSTR_MARK = tuple([f"{_h:}_mark" for _h in ATTR_REF])

    # constraints on the parameters
    D_CONSTRAINTS = {}

    # default values for the parameters
    D_DEFAULT = {}
    for key in ATTR_SIGMA:
        D_DEFAULT[key] = 0.
    for key in (ATTR_CONSTR_FLAG + ATTR_REF_FLAG):
        D_DEFAULT[key] = False
    for key in ATTR_CONSTR_MARK:
        D_DEFAULT[key] = ""

    PREFIX = "pd_instr_reflex_asymmetry"

    def __init__(self, **kwargs) -> NoReturn:
        super(PdInstrReflexAsymmetry, self).__init__()

        # defined for any integer and float parameters
        D_MIN = {"ttheta": 0}

        # defined for ani integer and float parameters
        D_MAX = {}

        self.__dict__["D_MIN"] = D_MIN
        self.__dict__["D_MAX"] = D_MAX
        for key, attr in self.D_DEFAULT.items():
            setattr(self, key, attr)
        for key, attr in kwargs.items():
            setattr(self, key, attr)

    def _func_fa(self, tth):
        """
        For assymmetry correction F_a(z).
        """
        return 2*tth*numpy.exp(-tth**2)

    def _func_fb(self, tth):
        """
        For assymmetry correction F_b(z)
        """
        return 2.*(2.*tth**2-3.)* self._func_fa(tth)

    def calc_asymmetry(self, tth, tth_hkl, fwhm):
        """
        Calculate asymmetry coefficients for  on the given list ttheta for
        bragg reflections flaced on the position ttheta_hkl
        tth and tth_hkl in degrees.

        look page 54 in FullProf Manual
        """
        tth_2d, tth_hkl_2d = numpy.meshgrid(tth, tth_hkl, indexing="ij")
        np_zero = numpy.zeros(tth_2d.shape, dtype = float)
        np_one = numpy.ones(tth_2d.shape, dtype = float)
        val_1, val_2 = np_zero, np_zero

        z_2d = (tth_2d - tth_hkl_2d)/fwhm[numpy.newaxis, :]

        p1, p2 = float(self.p1), float(self.p2)
        p3, p4 = float(self.p3), float(self.p4)
        flag_1, flag_2 = False, False
        if ((p1!= 0.)|(p3!= 0.)):
            flag_1 = True
            fa = self._func_fa(z_2d)
        if ((p2!= 0.)|(p4!= 0.)):
            flag_2 = True
            fb = self._func_fb(z_2d)

        flag_3, flag_4 = False, False
        if ((p1!= 0.)|(p2!= 0.)):
            if flag_1:
                val_1 += p1*fa
                flag_3 = True
            if flag_2:
                val_1 += p2*fb
                flag_3 = True
            if flag_3:
                c1 = 1./numpy.tanh(0.5*tth_hkl)
                val_1 *= c1[numpy.newaxis, :]

        if ((p3!= 0.)|(p4!= 0.)):
            if flag_1:
                val_2 += p3*fa
                flag_4 = True
            if flag_2:
                val_2 += p4*fb
                flag_4 = True
            if flag_4:
                c2 = 1./numpy.tanh(tth_hkl)
                val_2 *= c2[numpy.newaxis, :]

        asymmetry_2d = np_one+val_1+val_2
        return asymmetry_2d

    def get_dictionary(self):
        res = {}
        res["asymmetry_parameters"] = numpy.array([
            self.p1, self.p2, self.p3,
            self.p4], dtype=float)

        res["flags_asymmetry_parameters"] = numpy.array([
            self.p1_refinement, self.p2_refinement, self.p3_refinement,
            self.p4_refinement], dtype=bool)
        return res

class PdInstrReflexAsymmetryL(LoopN):
    """Asymmetry of Bragg reflections for several 1d powder diffractometers.
    """
    ITEM_CLASS = PdInstrReflexAsymmetry
    ATTR_INDEX = None
    def __init__(self, loop_name: str = None, **kwargs) -> NoReturn:
        super(PdInstrReflexAsymmetryL, self).__init__()
        self.__dict__["items"] = form_items_by_dictionary(self.ITEM_CLASS, kwargs)
        self.__dict__["loop_name"] = loop_name


# s_cont = """
#  loop_
#  _pd_instr_reflex_asymmetry_p1
#  _pd_instr_reflex_asymmetry_p2
#  _pd_instr_reflex_asymmetry_p3
#  _pd_instr_reflex_asymmetry_p4
#  0.0 0.0 0.0 0.0
#  1.0 0.1 0.5 0.17(78)
# """

# obj = PdInstrReflexAsymmetryL.from_cif(s_cont)
# print(obj, end="\n\n")

